/*
 * Parts taken from the "exif.js" file, so that it can be used for XMP of e.g. PNG files;
 * The license is alike of the "exif.js" file (it is code taken from it), see the "LICENSE.md" file;
 */

var xmp_xmlHead = ""
    + 'xmlns:Iptc4xmpCore="http://iptc.org/std/Iptc4xmpCore/1.0/xmlns/" '
    + 'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
    + 'xmlns:tiff="http://ns.adobe.com/tiff/1.0/" '
    + 'xmlns:plus="http://schemas.android.com/apk/lib/com.google.android.gms.plus" '
    + 'xmlns:ext="http://www.gettyimages.com/xsltExtension/1.0" '
    + 'xmlns:exif="http://ns.adobe.com/exif/1.0/" '
    + 'xmlns:stEvt="http://ns.adobe.com/xap/1.0/sType/ResourceEvent#" '
    + 'xmlns:stRef="http://ns.adobe.com/xap/1.0/sType/ResourceRef#" '
    + 'xmlns:crs="http://ns.adobe.com/camera-raw-settings/1.0/" '
    + 'xmlns:xapGImg="http://ns.adobe.com/xap/1.0/g/img/" '
    + 'xmlns:Iptc4xmpExt="http://iptc.org/std/Iptc4xmpExt/2008-02-29/" ';

var xmp_xml2json = function (xml) {
    var json = {};

    if (xml.nodeType == 1) { // element node
      if (xml.attributes.length > 0) {
        json['@attributes'] = {};
        for (var j = 0; j < xml.attributes.length; j++) {
          var attribute = xml.attributes.item(j);
          json['@attributes'][attribute.nodeName] = attribute.nodeValue;
        }
      }
    } else if (xml.nodeType == 3) { // text node
      return xml.nodeValue;
    }

    // deal with children
    if (xml.hasChildNodes()) {
      for(var i = 0; i < xml.childNodes.length; i++) {
        var child = xml.childNodes.item(i);
        var nodeName = child.nodeName;
        if (json[nodeName] == null) {
          json[nodeName] = xmp_xml2json(child);
        } else {
          if (json[nodeName].push == null) {
            var old = json[nodeName];
            json[nodeName] = [];
            json[nodeName].push(old);
          }
          json[nodeName].push(xmp_xml2json(child));
        }
      }
    }

    return json;
};

var xmp_xml2Object = function (xml) {
  try {
      var obj = {};
      if (xml.children.length > 0) {
        for (var i = 0; i < xml.children.length; i++) {
          var item = xml.children.item(i);
          var attributes = item.attributes;
          for(var idx in attributes) {
              var itemAtt = attributes[idx];
              var dataKey = itemAtt.nodeName;
              var dataValue = itemAtt.nodeValue;

              if(dataKey !== undefined) {
                  obj[dataKey] = dataValue;
              }
          }
          var nodeName = item.nodeName;

          if (typeof (obj[nodeName]) == "undefined") {
            obj[nodeName] = xmp_xml2json(item);
          } else {
            if (typeof (obj[nodeName].push) == "undefined") {
              var old = obj[nodeName];

              obj[nodeName] = [];
              obj[nodeName].push(old);
            }
            obj[nodeName].push(xmp_xml2json(item));
          }
        }
      } else {
        obj = xml.textContent;
      }
      return obj;
    } catch (e) {
        return {};
    }
};

