////////////////////////////////////////////////////////////////////////
// pca-wise methods

var pca_compute_cross_product = function (vector_1, vector_2) {
  debug_output("pca_compute_cross_product");

  var res_x = vector_1[1]*vector_2[2] - vector_1[2]*vector_2[1];
  var res_y = vector_1[2]*vector_2[0] - vector_1[0]*vector_2[2];
  var res_z = vector_1[0]*vector_2[1] - vector_1[1]*vector_2[0];

  if ((res_x + res_y + res_z) < 0) {
    res_x *= -1;
    res_y *= -1;
    res_z *= -1;
  }

  const size_all = Math.sqrt((res_x**2) + (res_y**2) + (res_z**2));
  if (size_all == 0) {
    // this should not happen, though zero 3rd principal component is acceptable
    return [0, 0, 0];
  }
  return [res_x/size_all, res_y/size_all, res_z/size_all];
};

var pca_normalize_one_eigen_vector = function (res_x, res_y, res_z) {
  debug_output("pca_normalize_one_eigen_vector");

  const size_all = Math.sqrt((res_x**2) + (res_y**2) + (res_z**2));
  if (size_all==0) {
    debug_output("PCA: zero size of a given eigen-vector zero");
    return [];
  } else {
    res_x /= size_all;
    res_y /= size_all;
    res_z /= size_all;
  }

  if ((res_x + res_y + res_z) < 0) {
    res_x *= -1;
    res_y *= -1;
    res_z *= -1;
  }

  return [res_x, res_y, res_z];
};

var pca_compute_one_eigen_vector = function (row_1, row_2, row_3, prev_vector) {
  debug_output("pca_compute_one_eigen_vector");

  if (row_1[0]==0) {
    if (row_2[0]!=0) {
      let row_aux = row_1;
      row_1 = row_2;
      row_2 = row_aux;
    } else {
      let row_aux = row_1;
      row_1 = row_3;
      row_3 = row_aux;
    }
  }

  let val_x_undetermined = false;

  if (row_1[0]==0) {
    val_x_undetermined = true;
  }

  if (!val_x_undetermined) {
    if (row_2[0]!=0) {
      row_2 = [0, row_2[1]*row_1[0] - row_1[1]*row_2[0], row_2[2]*row_1[0] - row_1[2]*row_2[0]];
    }
    if (row_3[0]!=0) {
      row_3 = [0, row_3[1]*row_1[0] - row_1[1]*row_3[0], row_3[2]*row_1[0] - row_1[2]*row_3[0]];
    }
  }

  // basically:
  // y * (row_2[1]-row_3[1]) == z * (row_3[2]-row_2[2])
  // if (row_2[1]-row_3[1])!=0 then:
  // z:=1
  // y = (row_3[2]-row_2[2]) / (row_2[1]-row_3[1])
  // else: vice-versa;

  let val_y_part = row_2[1]-row_3[1];
  let val_z_part = row_3[2]-row_2[2];

  let res_x = 1;
  let res_y = 1;
  let res_z = 1;

  if ((val_y_part != 0) && (val_z_part != 0)) {
    // here res_z == 1 (before normalization)
    res_y = val_z_part / val_y_part;
    if (!val_x_undetermined) {
      res_x = (-1/row_1[0]) * (row_1[1]*res_y + row_1[2]*res_z);
    } else {
      if ((!prev_vector) || (prev_vector[0]==0)) {
        res_x = 0;
      } else {
        res_x = (-1/prev_vector[0]) * ((res_y*prev_vector[1]) + (res_z*prev_vector[2]));
      }
    }
    return pca_normalize_one_eigen_vector(res_x, res_y, res_z);
  }

  // sort-of exceptional cases below

  if ((val_y_part == 0) && (val_z_part == 0)) {
    if (!prev_vector) {
      if (!val_x_undetermined) {
        res_x = (-1/prev_vector[0]) * ((res_y*prev_vector[1]) + (res_z*prev_vector[2]));
        return pca_normalize_one_eigen_vector(res_x, res_y, res_z);
      } else {
        let sqr3inv = 1.0/Math.sqrt(3);
        return [sqr3inv, sqr3inv, sqr3inv];
      }
    } else {
      if (!val_x_undetermined) {
        // prev_vector[0] * (1/row_1[0]) * (row_1[1]*res_y + row_1[2]*res_z) = prev_vector[1]*res_y + prev_vector[2]*res_z;
        // ((row_1[1]/row_1[0])*prev_vector[0] - prev_vector[1]) * res_y = (prev_vector[2] - (row_1[2]/row_1[0])*prev_vector[0]) * res_z
        if ((row_1[1]/row_1[0])*prev_vector[0] != prev_vector[1]) {
          res_y = (prev_vector[2] - (row_1[2]/row_1[0])*prev_vector[0]) / ((row_1[1]/row_1[0])*prev_vector[0] - prev_vector[1]);
        } else if (prev_vector[2] != (row_1[2]/row_1[0])*prev_vector[0]) {
          res_z = ((row_1[1]/row_1[0])*prev_vector[0] - prev_vector[1]) / (prev_vector[2] - (row_1[2]/row_1[0])*prev_vector[0]);
        }
        res_x = (-1/row_1[0]) * (row_1[1]*res_y + row_1[2]*res_z);
        return pca_normalize_one_eigen_vector(res_x, res_y, res_z);
      } else {
        if ((prev_vector[0]==0) || (prev_vector[1]==0) || (prev_vector[2]==0)) {
          if (prev_vector[0]==0) {
            return [1,0,0];
          } else if (prev_vector[1]==0) {
            return [0,1,0];
          } else {
            return [0,0,1];
          }
        } else {
          res_x = (-1/prev_vector[0]) * (prev_vector[1] + prev_vector[2]);
          return pca_normalize_one_eigen_vector(res_x, res_y, res_z);
        }
      }
    }
  }

  // here exactly one from val_y_part, val_z_part is zero;
  // val_y_part!=0 means res_y==0, and alike for z;

  if (!val_x_undetermined) {
    /*
      considering prev_vector:
      // case for res_y == 0:
      // prev_vector[0] * (1/row_1[0]) * (row_1[2]*res_z) = prev_vector[2]*res_z;
      // prev_vector[0] * (row_1[2]/row_1[0]) = prev_vector[2];
      // case for res_z == 0:
      // prev_vector[0] * (1/row_1[0]) * (row_1[1]*res_y) = prev_vector[1]*res_y;
      // prev_vector[0] * (row_1[1]/row_1[0]) = prev_vector[1];
      // that is, the prev_vector does not help here;
    */
    if (val_y_part != 0) {
      res_y = 0;
    } else {
      res_z = 0;
    }
    res_x = (-1/row_1[0]) * (row_1[1]*res_y + row_1[2]*res_z);
    return pca_normalize_one_eigen_vector(res_x, res_y, res_z);
  }

  if (prev_vector) {
    /*
      considering prev_vector:
      // case for res_y == 0:
      // prev_vector[0] * res_x = -1 * prev_vector[2] * res_z;
      // case for res_z == 0:
      // prev_vector[0] * res_x = -1 * prev_vector[1] * res_y;
    */
    if (val_y_part != 0) {
      res_y = 0;
      if (prev_vector[0] != 0) {
        res_z = 1;
        res_x = -1 * prev_vector[2] / prev_vector[0];
      } else if (prev_vector[2] != 0) {
        res_x = 1;
        res_z = -1 * prev_vector[0] / prev_vector[2];
      } else {
        res_x = 1;
        res_z = 1;
      }
    } else {
      res_z = 0;
      if (prev_vector[0] != 0) {
        res_y = 1;
        res_x = -1 * prev_vector[1] / prev_vector[0];
      } else if (prev_vector[1] != 0) {
        res_x = 1;
        res_y = -1 * prev_vector[0] / prev_vector[1];
      } else {
        res_x = 1;
        res_y = 1;
      }
    }
    return pca_normalize_one_eigen_vector(res_x, res_y, res_z);
  }

  if (val_y_part != 0) {
    res_y = 0;
  } else {
    res_z = 0;
  }
  return pca_normalize_one_eigen_vector(res_x, res_y, res_z);
};

var pca_sort_eigen_values = function (val_x1_ini, val_x2_ini, val_x3_ini) {
  debug_output("pca_sort_eigen_values");

  if (Math.abs(val_x1_ini) < Math.abs(val_x2_ini)) {
    let val_aux = val_x1_ini;
    val_x1_ini = val_x2_ini;
    val_x2_ini = val_aux;
  }
  if (Math.abs(val_x1_ini) < Math.abs(val_x3_ini)) {
    let val_aux = val_x1_ini;
    val_x1_ini = val_x3_ini;
    val_x3_ini = val_aux;
  } else {
    if (Math.abs(val_x2_ini) < Math.abs(val_x3_ini)) {
      let val_aux = val_x2_ini;
      val_x2_ini = val_x3_ini;
      val_x3_ini = val_aux;
    }
  }

  return [val_x1_ini, val_x2_ini, val_x3_ini];
};

var pca_compute_eigen_values = function (var_rr, var_gg, var_bb, var_gb, var_br, var_rg) {
  debug_output("pca_compute_eigen_values");

/*
  evaluating the lambdas via setting zero-valued determinant

  | rr-l rg   br   |
  | rg   gg-l gb   | == 0
  | br   gb   bb-l |

  + [rr*gg*bb - l*(rr*gg+gg*bb+bb*rr) + (l**2)*(rr+gg+bb) - (l**3) + 2*rg*gb*br]
  - [rr*(gb**2) + gg*(br**2) + bb*(rg**2) - l*((gb**2) + (br**2) + (rg**2))]

  a3*(l**3) + a2*(l**2) + a1*l + a0 = 0

  a3 = -1
  a2 = rr+gg+bb
  a1 = (gb**2) + (br**2) + (rg**2) - (rr*gg+gg*bb+bb*rr)
  a0 = rr*gg*bb + 2*rg*gb*br - (rr*(gb**2) + gg*(br**2) + bb*(rg**2))

  x = y - a2/(3*a3) = y + a2/3

  (y**3) + 3*p*y + 2*q = 0
  3*p = (3*a3*a1 - (a2**2)) * 1/(3*a3**2)
      = (3*(-1)*a1 - (a2**2)) * 1/3
  2*q = (2*a2**3)/(27*a3**3) + (a0/a3) - (a2*a1)/(3*a3**2)
      = (2*a2**3)/(-27) + (-a0) - (a2*a1)/(3)

  p = (-1/9) * (3*a1 + (a2**2))
  q = (-1/2) * (2*(a2**3)/27 + a0 + (a2*a1)/3)

*/

  const val_a2 = var_rr + var_gg + var_bb;
  const val_a1 = (var_gb**2) + (var_br**2) + (var_rg**2) - ((var_rr*var_gg) + (var_gg*var_bb) + (var_bb*var_rr));
  const val_a0 = var_rr*var_gg*var_bb + 2*var_rg*var_gb*var_br - (var_rr*(var_gb**2) + var_gg*(var_br**2) + var_bb*(var_rg**2));

  const val_p = (-1/9) * (3*val_a1 + (val_a2**2));
  const val_q = (-1/2) * (2*(val_a2**3)/27 + val_a0 + (val_a2*val_a1)/3);

  const val_y_to_x = val_a2/3;

  if (val_q==0) {
    // simpler solving:
    // y: 0, +-sqrt(-3*p)
    if (p > 0) {
      debug_output("PCA: val_q==0, val_p>0, this would lead to complex roots");
      return [];
    }
    let m3p_sqrt = Math.sqrt(-3*p);
    return pca_sort_eigen_values(val_y_to_x+m3p_sqrt, val_y_to_x, val_y_to_x-m3p_sqrt);
  }
  if (val_p==0) {
    // simpler solving:
    // y: -((2*q)**(1/3))
    let m2q_3root = -((2*q)**(1.0/3.0));
    return pca_sort_eigen_values(val_y_to_x+m2q_3root, val_y_to_x+m2q_3root, val_y_to_x+m2q_3root);
  }

  if (val_p>0) {
    debug_output("PCA: val_p>0, this would lead to complex roots");
    return [];
  }
  if (((val_p**3) + (val_q**2)) > 0) {
    debug_output("PCA: val_p**3 + val_q**2 > 0, this would lead to complex roots");
    return [];
  }

  const val_r = (val_q>0) ? Math.sqrt(Math.abs(val_p)) : -Math.sqrt(Math.abs(val_p));
  const val_fi = Math.acos(val_q/(val_r**3));

  var val_x1_ini = -2*val_r*Math.cos(val_fi/3) + val_y_to_x;
  var val_x2_ini = 2*val_r*Math.cos((Math.PI-val_fi)/3) + val_y_to_x;
  var val_x3_ini = 2*val_r*Math.cos((Math.PI+val_fi)/3) + val_y_to_x;

  return pca_sort_eigen_values(val_x1_ini, val_x2_ini, val_x3_ini);
};

// using the covariance-matrix way
var pca_compute_principle_components = function (can_ori_ctx, can_width, can_height) {
  debug_output("pca_compute_principle_components");

  var ori_data = can_ori_ctx.getImageData(0, 0, can_width, can_height);
  var ori_arr = ori_data.data;
  var arr_len = ori_arr.length;

  if (arr_len==0) {
    debug_output("PCA: empty input data!");
    return {};
  }

  var mean_r = 0;
  var mean_g = 0;
  var mean_b = 0;
  var mean_sq_rr = 0;
  var mean_sq_gg = 0;
  var mean_sq_bb = 0;
  var mean_sq_rg = 0;
  var mean_sq_gb = 0;
  var mean_sq_br = 0;

  let idx = arr_len-1;
  while (idx>=0) {
    idx--;
    let val_b = ori_arr[idx];
    idx--;
    let val_g = ori_arr[idx];
    idx--;
    let val_r = ori_arr[idx];
    idx--;
    mean_r += val_r;
    mean_g += val_g;
    mean_b += val_b;
    mean_sq_rr += val_r*val_r;
    mean_sq_gg += val_g*val_g;
    mean_sq_bb += val_b*val_b;
    mean_sq_rg += val_r*val_g;
    mean_sq_gb += val_g*val_b;
    mean_sq_br += val_b*val_r;
  }

  const pix_size = can_width * can_height;

  mean_r /= pix_size;
  mean_g /= pix_size;
  mean_b /= pix_size;
  mean_sq_rr /= pix_size;
  mean_sq_gg /= pix_size;
  mean_sq_bb /= pix_size;
  mean_sq_rg /= pix_size;
  mean_sq_gb /= pix_size;
  mean_sq_br /= pix_size;

  const var_rr = mean_sq_rr - (mean_r*mean_r);
  const var_gg = mean_sq_gg - (mean_g*mean_g);
  const var_bb = mean_sq_bb - (mean_b*mean_b);
  const var_rg = mean_sq_rg - (mean_r*mean_g);
  const var_gb = mean_sq_gb - (mean_g*mean_b);
  const var_br = mean_sq_br - (mean_b*mean_r);

  // first to check whether it is (not) a grayscaled image, since there is no sense for PCA in such a case;
  if ((var_rr==var_gg) && (var_gg==var_bb) && (var_rg==var_gb) && (var_gb==var_br)) {
    debug_output("PCA: apparently grayscaled image, PCA not appropriate here");
    return {};
  }

  // if not grayscaled, but independent, just taking the colors (if anything in there)
  if ((var_rg == 0) && (var_gb == 0) && (var_br == 0)) {
    if ((((var_rr>0)?1:0) + ((var_gg>0)?1:0) + ((var_bb>0)?1:0)) < 2) {
      debug_output("PCA: apparently single-color image, PCA not appropriate here");
      return {};
    }
    let color_1 = [1,0,0];
    let color_2 = [0,1,0];
    let color_3 = [0,0,1];
    if (var_gg < var_bb) {
      color_2 = [0,0,1];
      color_3 = [0,1,0];
    }
    if (var_rr < ((var_gg*color_2[1])+(var_bb*color_2[2]))) {
      color_1 = color_2;
      color_2 = [1,0,0];
      if (var_rr < ((var_gg*color_3[1])+(var_bb*color_3[2]))) {
        color_2 = color_3;
        color_3 = [1,0,0];
      }
    }
    return {"mean_values":[mean_r, mean_g, mean_b], "eigen_vectors":[color_1, color_2, color_3]};
  }

  // dealing with the other cases as if they are general cases

/*
  schematic view on the evaluation, computing eigen-values first:

  /        \
  |rr rg br|    | x1 |       | x1 |
  |rg gg gb| *  | x2 | = l * | x2 |
  |br gb bb|    | x3 |       | x3 |
  \        /

 */

  let eigen_values = pca_compute_eigen_values(var_rr, var_gg, var_bb, var_gb, var_br, var_rg);
  if ((!eigen_values) || (eigen_values.length==0)) {
    debug_output("PCA: could not compute eigen values!");
    return {};
  }
  const val_x1 = eigen_values[0];
  const val_x2 = eigen_values[1];
  const val_x3 = eigen_values[2];

  debug_output("PCA: val_x1: " + val_x1 + ", val_x2: " + val_x2 + ", val_x3: " + val_x3);

  const var_rr_ml1 = var_rr - val_x1;
  const var_gg_ml1 = var_gg - val_x1;
  const var_bb_ml1 = var_bb - val_x1;

  let row_1 = [var_rr_ml1, var_rg, var_br];
  let row_2 = [var_rg, var_gg_ml1, var_gb];
  let row_3 = [var_br, var_gb, var_bb_ml1];

  const res_vec_1 = pca_compute_one_eigen_vector(row_1, row_2, row_3, null);
  if (res_vec_1.length == 0) {
    debug_output("PCA: could not compute the 1st eigen vector!");
    return {};
  }

  const var_rr_ml2 = var_rr - val_x2;
  const var_gg_ml2 = var_gg - val_x2;
  const var_bb_ml2 = var_bb - val_x2;

  row_1 = [var_rr_ml2, var_rg, var_br];
  row_2 = [var_rg, var_gg_ml2, var_gb];
  row_3 = [var_br, var_gb, var_bb_ml2];

  const res_vec_2 = pca_compute_one_eigen_vector(row_1, row_2, row_3, res_vec_1);
  if (res_vec_2.length == 0) {
    debug_output("PCA: could not compute the 2nd eigen vector!");
    return {};
  }

  // it is simpler to compute the 3rd principal vector via cross-product,
  // since do not need to care about situations when colors are 2-dimensionally spread;
  const res_vec_3 = pca_compute_cross_product(res_vec_1, res_vec_2);
  if (res_vec_3.length == 0) {
    debug_output("PCA: could not compute the 3rd eigen vector!");
    return {};
  }

  debug_output([res_vec_1, res_vec_2, res_vec_3]);

  return {"mean_values":[mean_r, mean_g, mean_b], "eigen_vectors":[res_vec_1, res_vec_2, res_vec_3]};
};

var pca_fill_requested_projection = function (can_ori_ctx, can_res_ctx, can_width, can_height, color_means, eigen_vector, eigen_mode) {
  debug_output("pca_fill_requested_projection");

  var ori_data = can_ori_ctx.getImageData(0, 0, can_width, can_height);
  var ori_arr = ori_data.data;
  var arr_len = ori_arr.length;

  var res_data = can_res_ctx.getImageData(0, 0, can_width, can_height);
  var res_arr = res_data.data;

  if (arr_len != res_arr.length) {
    return;
  }
  if (arr_len==0) {
    return false;
  }

  if ((!color_means) || (color_means.length != 3)) {
    return false;
  }

  if ((!eigen_vector) || (eigen_vector.length != 3)) {
    return false;
  }

  const use_mean_r = color_means[0];
  const use_mean_g = color_means[1];
  const use_mean_b = color_means[2];

  const use_dir_r = eigen_vector[0];
  const use_dir_g = eigen_vector[1];
  const use_dir_b = eigen_vector[2];

  if (eigen_mode<0) {
    let idx_set = arr_len-1;
    while (idx_set>=0) {
      res_arr[idx_set] = 255;
      idx_set--;
      let val_b = ori_arr[idx_set] - use_mean_b;
      idx_set--;
      let val_g = ori_arr[idx_set] - use_mean_g;
      idx_set--;
      let val_r = ori_arr[idx_set] - use_mean_r;

      let use_proj = 127-(use_dir_r*val_r + use_dir_g*val_g + use_dir_b*val_b);
      res_arr[idx_set+2] = use_proj;
      res_arr[idx_set+1] = use_proj;
      res_arr[idx_set] = use_proj;
      idx_set--;
    }
  }
  if (eigen_mode>0) {
    let idx_set = arr_len-1;
    while (idx_set>=0) {
      res_arr[idx_set] = 255;
      idx_set--;
      let val_b = ori_arr[idx_set] - use_mean_b;
      idx_set--;
      let val_g = ori_arr[idx_set] - use_mean_g;
      idx_set--;
      let val_r = ori_arr[idx_set] - use_mean_r;

      let use_proj = 127+(use_dir_r*val_r + use_dir_g*val_g + use_dir_b*val_b);
      res_arr[idx_set+2] = use_proj;
      res_arr[idx_set+1] = use_proj;
      res_arr[idx_set] = use_proj;
      idx_set--;
    }
  }
  if (eigen_mode==0) {
    let idx_set = arr_len-1;
    while (idx_set>=0) {
      res_arr[idx_set] = 255;
      idx_set--;
      let val_b = ori_arr[idx_set] - use_mean_b;
      idx_set--;
      let val_g = ori_arr[idx_set] - use_mean_g;
      idx_set--;
      let val_r = ori_arr[idx_set] - use_mean_r;

      let use_proj = Math.abs(use_dir_r*val_r + use_dir_g*val_g + use_dir_b*val_b);
      res_arr[idx_set+2] = use_proj;
      res_arr[idx_set+1] = use_proj;
      res_arr[idx_set] = use_proj;
      idx_set--;
    }
  }

  can_res_ctx.putImageData(res_data, 0, 0);
};

