////////////////////////////////////////////////////////////////////////
// enabling/disabling UI parts according to (not) loaded image

var set_image_is_loaded = function (value) {
  debug_output("set_image_is_loaded");

  image_is_loaded = value;

  let bt_main_dis_class = "bt-main-disabled";

  let set_els = [
    document.getElementById("bt-main-levels"),
    document.getElementById("bt-main-slopes"),
    document.getElementById("bt-main-ela"),
    document.getElementById("bt-main-pca")
  ];

  for (let one_el_idx = set_els.length-1; one_el_idx >= 0; one_el_idx--) {
    let one_el = set_els[one_el_idx];
    if (one_el) {
      if (image_is_loaded) {
        if (one_el.classList.contains(bt_main_dis_class)) {
          one_el.classList.remove(bt_main_dis_class);
        }
      } else {
        if (!(one_el.classList.contains(bt_main_dis_class))) {
          one_el.classList.add(bt_main_dis_class);
        }
      }
    }
  }

  var notice_el = document.getElementById("field-meta-file-notice");
  var notice_without_image_class = "notice-without-opened-file";

  if (image_is_loaded) {
    if (bt_meta_exif) {
      if (image_has_exif) {
        bt_meta_exif.title = "";
      } else {
        bt_meta_exif.title = "No exif found at the image";
      }
    }
    if (bt_meta_thumb) {
      if (image_has_thumb) {
        bt_meta_thumb.title = "";
      } else {
        bt_meta_thumb.title = "No thumbnail found at the image";
      }
    }

    if (notice_el) {
      notice_el.onclick = null;
      if (notice_el.classList.contains(notice_without_image_class)) {
        notice_el.classList.remove(notice_without_image_class);
      }
    }

    return;
  }

  if (notice_el) {
    notice_el.onclick = function (evt) {
      blink_load_button();
    };
    if (!(notice_el.classList.contains(notice_without_image_class))) {
      notice_el.classList.add(notice_without_image_class);
    }
  }

  if (bt_meta_exif) {
    bt_meta_exif.title = "";
  }
  if (bt_meta_thumb) {
    bt_meta_thumb.title = "";
  }

  let inf_width = document.getElementById("info-sizes-width");
  let inf_height = document.getElementById("info-sizes-height");
  if (inf_width) {
    inf_width.innerHTML = "0";
  }
  if (inf_height) {
    inf_height.innerHTML = "0";
  }

  disable_exif();
  disable_thumb();

  let hid_els = [
    document.getElementById("field-meta-file-url-holder"),
    document.getElementById("field-meta-file-name-holder"),
    document.getElementById("field-meta-file-headers-holder"),
    document.getElementById("field-meta-file-start-holder")
  ];

  for (let one_el_idx = hid_els.length-1; one_el_idx >= 0; one_el_idx--) {
    let one_el = hid_els[one_el_idx];
    if (one_el) {
      if (!(one_el.classList.contains(class_el_hidden))) {
        one_el.classList.add(class_el_hidden);
      }
    }
  }

  image_file_start = "";
};

////////////////////////////////////////////////////////////////////////
// meta-related

var image_file_extract_comment_pngxmp = function (image_data, image_type) {
  debug_output("image_file_extract_comment_pngxmp");

  // since yet not having a function for dealing with PNG tags,
  // looking for "XMP inside PNG" here (along with the PNG comment tag);
  png_xmp_loading = {};

  var comment_str = "";
  var correct_text_necessary = false;

  let is_jpeg_like = false;
  let is_png_like = false;
  for (let img_type_part in {"jpeg":true, "jpg":true, "tiff":true, "tif":true}) {
    if (image_type.toLowerCase().indexOf(img_type_part)>-1) {
      is_jpeg_like = true;
    }
  }
  for (let img_type_part in {"png":true}) {
    if (image_type.toLowerCase().indexOf(img_type_part)>-1) {
      is_png_like = true;
    }
  }
  let max_search_start_use = 10000;
  if (is_png_like) {
    max_search_start_use = 100000;
  }
  let max_comment_start = image_data.length - 4;
  if (max_comment_start > max_search_start_use) {
    max_comment_start = max_search_start_use;
  }

  if (is_jpeg_like) {
    // leaving starting 6 bytes for the header itself,
    // since it may contain the 0xFF 0xFE byte pair;
    for (let ind_ini=6; ind_ini<max_comment_start; ind_ini++) {
      if (image_data.charCodeAt(ind_ini)==255) {
        if (image_data.charCodeAt(ind_ini+1)==254) {
          let comment_length = 256*image_data.charCodeAt(ind_ini+2)+image_data.charCodeAt(ind_ini+3);
          comment_str = image_data.substring(ind_ini+4, ind_ini+2+comment_length);
          if (ind_ini>2000) {
            correct_text_necessary = true;
          }
          break;
        }
      }
    }
  }

  if (is_png_like) {
    let comment_head_part = "Comment";
    let comment_head_part_len = comment_head_part.length;
    let comment_head = "tEXt" + comment_head_part;
    let comment_parts = image_data.split(comment_head);
    if ((comment_parts.length > 1) && (comment_parts[0].length > 4) && (comment_parts[1].length > 1) && (comment_parts[1].charCodeAt(0)==0)) {
      let first_part_len = comment_parts[0].length;
      // deliberately not considering huge comments, i.e. taking the first two bytes (of the length value) as zeros;
      let comment_length = comment_parts[0].charCodeAt(first_part_len-1) + (256*comment_parts[0].charCodeAt(first_part_len-2));
      if (comment_length > comment_head_part_len) {
        comment_str = comment_parts[1].substring(1, comment_length-comment_head_part_len);
      }
    }
  }

  if (is_png_like) {
    let xmp_str = "";
    let xmp_head_part = "XML:com.adobe.xmp";
    let xmp_head_part_len = xmp_head_part.length;
    let xmp_head = "iTXt" + xmp_head_part;
    let xmp_parts = image_data.split(xmp_head);
    if ((xmp_parts.length > 1) && (xmp_parts[0].length > 4) && (xmp_parts[1].length > 1) && (xmp_parts[1].charCodeAt(0)==0)) {
      let first_part_len = xmp_parts[0].length;
      // deliberately not considering huge XMP, i.e. taking the first two bytes (of the length value) as zeros;
      let xmp_length = xmp_parts[0].charCodeAt(first_part_len-1) + (256*xmp_parts[0].charCodeAt(first_part_len-2));
      if (xmp_length > xmp_head_part_len) {
        xmp_str = xmp_parts[1].substring(1, xmp_length-xmp_head_part_len);
      }
    }
    if (xmp_str != "") {
      while (xmp_str.charCodeAt(0) == 0) {
        xmp_str = xmp_str.slice(1);
      }
    }
    if (xmp_str != "") {
      let xmp_meta_start = xmp_str.indexOf("<x:xmpmeta");
      if (xmp_meta_start > -1) {
        xmp_meta_start += 10;
        if (xmp_meta_start < xmp_str.length) {
          xmp_str = xmp_str.slice(0, xmp_meta_start) + " " + xmp_xmlHead + " " + xmp_str.slice(xmp_meta_start);
        } else {
          xmp_str = "";
        }
      } else {
        xmp_str = "";
      }
    }
    if (xmp_str != "") {
      let xmp_parser = new DOMParser();;
      let xmp_obj = xmp_parser.parseFromString(xmp_str, 'text/xml');
      if (xmp_obj.getElementsByTagName("parsererror").length) {
        debug_output("could not parse the XMP from PNG");
      } else {
        try {
          xmp_obj = xmp_xml2Object(xmp_obj);
          png_xmp_loading = xmp_obj;
        } catch (e) {
          debug_output("could not transform XML created from PNG-wise XMP");
          png_xmp_loading = {};
        }
      }
    }
  }

  if (comment_str != "") {
    // first trying to put it into UTF-8
    let comment_str_utf8 = "";
    try {
      comment_str_utf8 = decodeURIComponent(escape(comment_str));
    } catch (e) {
      debug_output("comment incorrect");
      comment_str_utf8 = "";
    }
    if ((comment_str_utf8 != "") || correct_text_necessary) {
      comment_str = comment_str_utf8;
    }

    comment_str = safe_text_outer(comment_str).split(/\r\n|\r|\n/g).join("<br>");
  }

  return comment_str;
};

var guess_image_type = function (start_part, is_base64) {
  debug_output("guess_image_type");

  var guessed_type = "";

  try {
    let img_header = "";
    if (is_base64) {
      let use_length = start_part.length - (start_part.length % 4);
      if (use_length > 8) {
        use_length = 8;
      }
      start_part = window.atob(start_part.substring(0, use_length));
    } else {
      start_part = start_part.substring(0, 8);
    }
    if (start_part.length >= 3) {
      let start_arr = new Uint8Array(start_part.length);
      for (let arr_ind = start_part.length - 1; arr_ind >= 0; arr_ind--) {
        start_arr[arr_ind] = start_part.charCodeAt(arr_ind);
      }
      for (let arr_ind = 0; arr_ind < 3; arr_ind++) {
        img_header += start_arr[arr_ind].toString(16);
      }
      // https://mimesniff.spec.whatwg.org/#matching-an-image-type-pattern
      switch (img_header.toLowerCase()) {
        case "ffd8ff":
          guessed_type = "image/jpeg";
          break
        case "89504e":
          guessed_type = "image/png";
          break
        case "474946":
          guessed_type = "image/gif";
          break
        case "524946":
          guessed_type = "image/webp";
          break
        default:
          guessed_type = "";
      }
    }
  } catch (e) {
    guessed_type = "";
  }
  return guessed_type;
};

var disable_exif = function () {
  debug_output("disable_exif");

  image_has_exif = false;

  if (bt_meta_exif && (!(bt_meta_exif.classList.contains(meta_section_not)))) {
    bt_meta_exif.classList.add(meta_section_not);
  }
};

var disable_thumb = function () {
  debug_output("disable_thumb");

  image_has_thumb = false;

  if (bt_meta_thumb && (!(bt_meta_thumb.classList.contains(meta_section_not)))) {
    bt_meta_thumb.classList.add(meta_section_not);
  }
};

var inner_trim_from_meta_array = function (input) {
  debug_output("inner_trim_from_meta_array");

  var output = [];
  var use_trimmed = true;
  try {
    let idx_end = input.length-1;
    for (; idx_end>=0;) {
      if (input[idx_end]!=0) {
        break;
      }
      idx_end--;
    }
    if (idx_end>=0) {
      output = Array(idx_end+1);
    }
    for (let idx_inn=idx_end; idx_inn>=0; idx_inn--) {
      output[idx_inn] = input[idx_inn];
    }
  } catch (e) {
    use_trimmed = false;
  }

  if (!use_trimmed) {
    return input;
  }
  return output;
};

var inner_empty_from_meta_array = function (input) {
  debug_output("inner_empty_from_meta_array");

  var pos_first = 0;
  try {
    var starts_with_label = true;
    var input_len = input.length;
    // these say: "ASCII", "UNICODE", filled by zeros to length 8;
    var onset_parts = [[65, 83, 67, 73, 73, 0, 0, 0], [85, 78, 73, 67, 79, 68, 69, 0]];
    var onset_parts_len = onset_parts.length;
    var onset_len = 8;
    starts_with_label = false;
    if (input_len >= onset_len) {
      for (let idx_par = onset_parts_len-1; idx_par>=0; idx_par--) {
        let onset = onset_parts[idx_par];
        let starts_with_part_label = true;
        for (let idx_ons = onset_len-1; idx_ons>=0; idx_ons--) {
          if (input[idx_ons] != onset[idx_ons]) {
            starts_with_part_label = false;
            break;
          }
        }
        if (starts_with_part_label) {
          starts_with_label = true;
          break;
        }
      }
    }
    if (starts_with_label) {
      pos_first = onset_len;
    }
  } catch (e) {
    pos_first = 0;
  }

  try {
    for (let idx=input.length-1; idx>=pos_first; idx--) {
      if (!(input[idx] in {0:true, "0":true, 32:true, "32":true})) {
        return false;
      }
    }
  } catch (e) {
    return false;
  }

  return true;
};

var inner_unicode_from_meta_array = function (input) {
  debug_output("inner_unicode_from_meta_array");

  var input_len = input.length;
  // this says: "UNICODE", filled by zeros to length 8;
  var onset = [85, 78, 73, 67, 79, 68, 69, 0];
  var onset_len = onset.length;
  if (input_len <= onset_len) {
    return "";
  }
  for (let idx_ons = onset_len-1; idx_ons>=0; idx_ons--) {
    if (input[idx_ons] != onset[idx_ons]) {
      return "";
    }
  }
  if (((input_len - onset_len) % 2) != 0) {
    return "";
  }

  var res = "";
  for (let idx_res = onset_len; idx_res<input_len; idx_res=idx_res+2) {
    let res_part_1 = input[idx_res];
    if (typeof res_part_1 !== "number") {
      res_part_1 = parseInt(res_part_1);
      if (!isFinite(res_part_1)) {
        return "";
      }
    }
    let res_part_2 = input[idx_res+1];
    if (typeof res_part_2 !== "number") {
      res_part_2 = parseInt(res_part_2);
      if (!isFinite(res_part_2)) {
        return "";
      }
    }

    res += "&#" + ((256*res_part_2) + res_part_1) + ";";
  }

  return res;
};

var inner_ascii_from_meta_array = function (input) {
  debug_output("inner_ascii_from_meta_array");

  var input_len = input.length;
  // this says: "ASCII", filled by zeros to length 8;
  var onset = [65, 83, 67, 73, 73, 0, 0, 0];
  var onset_len = onset.length;
  if (input_len <= onset_len) {
    return "";
  }
  for (let idx_ons = onset_len-1; idx_ons>=0; idx_ons--) {
    if (input[idx_ons] != onset[idx_ons]) {
      return "";
    }
  }

  var res = "";
  for (let idx_res = onset_len; idx_res<input_len; idx_res=idx_res+1) {
    let res_part_1 = input[idx_res];
    if (typeof res_part_1 !== "number") {
      res_part_1 = parseInt(res_part_1);
      if (!isFinite(res_part_1)) {
        return "";
      }
    }

    res += "&#" + (res_part_1) + ";";
  }

  return res;
};

var inner_string_from_meta_array = function (input) {
  debug_output("inner_string_from_meta_array");

  var str_part = "";
  var is_whole = true;
  var in_taking = true;
  var idx = -1;
  try {
    while (true) {
      idx += 1;
      if (typeof input[idx] === "number") {
        if (input[idx] == 0) {
          in_taking = false;
          continue;
        }
        if (!in_taking) {
          is_whole = false;
          break;
        }
        let value = input[idx];
        str_part += String.fromCharCode(value);
        continue;
      }
      break;
    }
    if (str_part != "") {
      str_part = decodeURIComponent(escape(str_part));
    }
  } catch (e) {
    str_part = "";
  }

  if (str_part != "") {
    if (!is_whole) {
      str_part += " ...";
    }
    return safe_text_inner(str_part);
  }

  is_whole = true;
  in_taking = true;
  idx = -1;
  while (true) {
    idx += 1;
    if (typeof input[idx] === "number") {
      let value = input[idx];
      if (in_taking) {
        if ((value==10) || (value==13) || (value==9)) {
          str_part += " ";
          continue;
        }
        if ((value>31) && (value<127)) {
          str_part += "&#" + value + ";";
          continue;
        }
        if (value==0) {
          in_taking = false;
          continue;
        }
        in_taking = false;
        is_whole = false;
        break;
      }
      if (value==0) {
        continue;
      }
      is_whole = false;
      break;
    } else {
      is_whole = false;
      break;
    }
  }

  var to_cont = false;
  if ((str_part != "") && (!is_whole)) {
    str_part += " ...";
  }

  return str_part;
};

var show_image_base = function (resp_headers) {
  debug_output("show_image_base");

  last_loaded_path = last_loading_path;
  last_loaded_local = last_loading_local;

  let comment_holder_el = document.getElementById("field-meta-file-comment-holder");
  let comment_el = document.getElementById("field-meta-file-comment");
  if (comment_el) {
    comment_el.innerHTML = image_file_comment;
  }
  if (comment_holder_el) {
    if (image_file_comment != "") {
      if (comment_holder_el.classList.contains(class_el_hidden)) {
        comment_holder_el.classList.remove(class_el_hidden);
      }
    } else {
      if (!(comment_holder_el.classList.contains(class_el_hidden))) {
        comment_holder_el.classList.add(class_el_hidden);
      }
    }
  }

  if (!resp_headers) {
    resp_headers = [];
  }
  var resp_headers_len = resp_headers.length;

  var headers_content = "";
  var headers_got_some = false;
  for (let one_resp_part_idx = 0; one_resp_part_idx < resp_headers_len; one_resp_part_idx++) {
    let one_resp_header = resp_headers[one_resp_part_idx];
    let one_resp_header_key = one_resp_header[0];
    let one_resp_header_value = one_resp_header[1];
    if (one_resp_header_key.toLowerCase() == "last-modified") {
      headers_content += "<tr><td class=\"" + class_header_key + " " + class_header_highlight + "\">";
      headers_content += safe_text_outer(one_resp_header_key);
      headers_content += "</td><td class=\"" + class_header_highlight + "\">";
      headers_content += safe_text_outer(one_resp_header_value);
    } else {
      headers_content += "<tr><td class=\"" + class_header_key + "\">";
      headers_content += safe_text_outer(one_resp_header_key);
      headers_content += "</td><td>";
      headers_content += safe_text_outer(one_resp_header_value);
    }
    headers_content += "</td></tr>";
    headers_got_some = true;
  }

  var headers_el = document.getElementById("field-meta-file-headers");
  if (headers_el) {
    headers_el.innerHTML = headers_content;
  }

  var headers_hld_el = document.getElementById("field-meta-file-headers-holder");
  if (headers_hld_el) {
    if (headers_got_some) {
      if (headers_hld_el.classList.contains(class_el_hidden)) {
        headers_hld_el.classList.remove(class_el_hidden);
      }
    } else {
      if (!(headers_hld_el.classList.contains(class_el_hidden))) {
        headers_hld_el.classList.add(class_el_hidden);
      }
    }
  }

  var url_el = document.getElementById("field-meta-file-url");
  var url_hld_el = document.getElementById("field-meta-file-url-holder");
  var fname_el = document.getElementById("field-meta-file-name");
  var fname_hld_el = document.getElementById("field-meta-file-name-holder");

  if (url_el) {
    if (last_loaded_local) {
      url_el.href = "";
      url_el.innerHTML = "";
    } else {
      url_el.href = last_loaded_path;
      url_el.innerHTML = safe_text_outer(last_loaded_path);
    }
  }
  if (url_hld_el) {
    if ((!last_loaded_local) && (last_loaded_path!="")) {
      if (url_hld_el.classList.contains(class_el_hidden)) {
        url_hld_el.classList.remove(class_el_hidden);
      }
    } else {
      if (!(url_hld_el.classList.contains(class_el_hidden))) {
        url_hld_el.classList.add(class_el_hidden);
      }
    }
  }

  if (fname_el) {
    if (last_loaded_local) {
      fname_el.innerHTML = safe_text_outer(last_loaded_path);
    } else {
      fname_el.innerHTML = "";
    }
  }
  if (fname_hld_el) {
    if (last_loaded_local && (last_loaded_path!="")) {
      if (fname_hld_el.classList.contains(class_el_hidden)) {
        fname_hld_el.classList.remove(class_el_hidden);
      }
    } else {
      if (!(fname_hld_el.classList.contains(class_el_hidden))) {
        fname_hld_el.classList.add(class_el_hidden);
      }
    }
  }

  var start_holder_el = document.getElementById("field-meta-file-start-holder");
  var start_show_el = document.getElementById("field-meta-file-start-show");
  var start_part_el = document.getElementById("field-meta-file-start-part");
  var start_hide_el = document.getElementById("field-meta-file-start-hide");

  let class_part_less_one = class_part_less + "-1";
  let class_part_more_one = class_part_more + "-1";

  if (start_holder_el) {
    if (start_holder_el.classList.contains(class_el_hidden)) {
      start_holder_el.classList.remove(class_el_hidden);
    }
  }
  if (start_show_el) {
    if (start_show_el.classList.contains(class_el_hidden)) {
      start_show_el.classList.remove(class_el_hidden);
    }
    if (!(start_show_el.classList.contains(class_part_less_one))) {
      start_show_el.classList.add(class_part_less_one);
    }
    start_show_el.innerHTML = "<span id=\"id-" + class_show_more + "-1\" class=\"" + class_show_more + " bt-show-more-less without-selection\" tabIndex=\"1\">show</span>";
  }
  if (start_part_el) {
    if (!(start_part_el.classList.contains(class_el_hidden))) {
      start_part_el.classList.add(class_el_hidden);
    }
    if (!(start_part_el.classList.contains(class_part_more_one))) {
      start_part_el.classList.add(class_part_more_one);
    }
    start_part_el.innerHTML = image_file_start;
  }
  if (start_hide_el) {
    if (!(start_hide_el.classList.contains(class_el_hidden))) {
      start_hide_el.classList.add(class_el_hidden);
    }
    if (!(start_hide_el.classList.contains(class_part_more_one))) {
      start_hide_el.classList.add(class_part_more_one);
    }
    start_hide_el.innerHTML = "<span id=\"id-" + class_show_less + "-1\" class=\"" + class_show_less + " bt-show-more-less without-selection\" tabIndex=\"1\">hide</span>";
  }
};

////////////////////////////////////////////////////////////////////////
// notices, on image loading/preparing

var show_image_notice = function (force) {
  debug_output("show_image_notice");

  if (to_hide_notice) {
    clearTimeout(to_hide_notice);
    to_hide_notice = null;
  }

  let notice_hld_el = document.getElementById("field-meta-file-notice-holder");
  if (notice_hld_el) {
    if (notice_hld_el.classList.contains(class_el_hidden)) {
      notice_hld_el.classList.remove(class_el_hidden);
    }
  }

  if (force) {
    disp_part_meta = "file";
    main_use_part("meta");
  }
};

var hide_image_notice = function (timeout) {
  debug_output("hide_image_notice");

  if (to_hide_notice) {
    clearTimeout(to_hide_notice);
    to_hide_notice = null;
  }
  if (typeof timeout !== "undefined") {
    if (timeout > 0) {
      to_hide_notice = setTimeout(function () {
        hide_image_notice();
      }, timeout);
      return;
    }
  }

  let notice_hld_el = document.getElementById("field-meta-file-notice-holder");
  if (notice_hld_el) {
    if (!(notice_hld_el.classList.contains(class_el_hidden))) {
      notice_hld_el.classList.add(class_el_hidden);
    }
  }
};

var set_image_notice = function (notice) {
  debug_output("set_image_notice: " + notice);

  let notice_el = document.getElementById("field-meta-file-notice");
  if (notice_el) {
    notice_el.innerHTML = safe_text_outer(notice);
  }
};

////////////////////////////////////////////////////////////////////////
// if has not got image as a local file

var image_loaded_none = function (notice) {
  debug_output("image_loaded_none: " + notice);

  var use_notice = "No image file provided";
  if ((typeof notice !== "undefined") && notice && (notice != "")) {
    use_notice = notice;
  }
  set_image_notice(use_notice);
  show_image_notice(true);
  if (image_is_loaded) {
    hide_image_notice(default_notice_timeout);
  }
};

////////////////////////////////////////////////////////////////////////
// setting events for showing more/less parts where appropriate

var do_show_more_parts = function (use_rank) {
  debug_output("do_show_more_parts");

  let tags_to_hide = document.getElementsByClassName(class_part_less + "-" + use_rank);
  for (let ind = tags_to_hide.length - 1; ind>=0; ind--) {
    let hide_el = tags_to_hide[ind];
    if (!(hide_el.classList.contains(class_el_hidden))) {
      hide_el.classList.add(class_el_hidden);
    }
  }
  let tags_to_show = document.getElementsByClassName(class_part_more + "-" + use_rank);
  for (let ind = tags_to_show.length - 1; ind>=0; ind--) {
    let show_el = tags_to_show[ind];
    if (show_el.classList.contains(class_el_hidden)) {
      show_el.classList.remove(class_el_hidden);
    }
  }
};

var do_show_less_parts = function (use_rank) {
  debug_output("do_show_less_parts");

  let tags_to_hide = document.getElementsByClassName(class_part_more + "-" + use_rank);
  for (let ind = tags_to_hide.length - 1; ind>=0; ind--) {
    let hide_el = tags_to_hide[ind];
    if (!(hide_el.classList.contains(class_el_hidden))) {
      hide_el.classList.add(class_el_hidden);
    }
  }
  let tags_to_show = document.getElementsByClassName(class_part_less + "-" + use_rank);
  for (let ind = tags_to_show.length - 1; ind>=0; ind--) {
    let show_el = tags_to_show[ind];
    if (show_el.classList.contains(class_el_hidden)) {
      show_el.classList.remove(class_el_hidden);
    }
  }
};

var setup_show_more_less_parts = function () {
  debug_output("setup_show_more_less_parts");

  let tags_show_more = document.getElementsByClassName(class_show_more);
  let tags_show_more_len = tags_show_more.length;
  for (let ind = tags_show_more_len - 1; ind>=0; ind--) {
    let cur_el = tags_show_more[ind];
    let cur_id = cur_el.id;
    if (!cur_id) {
      continue;
    }
    let cur_rank_pos = cur_id.lastIndexOf("-") + 1;
    if (cur_rank_pos <= 0) {
      continue;
    }
    let cur_rank = cur_id.substring(cur_rank_pos);
    if (cur_rank == "") {
      continue;
    }

    cur_el.addEventListener("keydown", (function () {
      let use_rank = cur_rank;
      return function (evt) {
        evt = evt || window.event;
        if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
          prevent_default(evt);
          do_show_more_parts(use_rank);
        }
      };
    }) ());
    cur_el.addEventListener("click", (function () {
      let use_rank = cur_rank;
      return function (evt) {
        prevent_default(evt);
        do_show_more_parts(use_rank);
      };
    }) ());
  }

  let tags_show_less = document.getElementsByClassName(class_show_less);
  let tags_show_less_len = tags_show_less.length;
  for (let ind = tags_show_less_len - 1; ind>=0; ind--) {
    let cur_el = tags_show_less[ind];
    let cur_id = cur_el.id;
    if (!cur_id) {
      continue;
    }
    let cur_rank_pos = cur_id.lastIndexOf("-") + 1;
    if (cur_rank_pos <= 0) {
      continue;
    }
    let cur_rank = cur_id.substring(cur_rank_pos);
    if (cur_rank == "") {
      continue;
    }

    cur_el.addEventListener("keydown", (function () {
      let use_rank = cur_rank;
      return function (evt) {
        evt = evt || window.event;
        if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
          prevent_default(evt);
          do_show_less_parts(use_rank);
        }
      };
    }) ());
    cur_el.addEventListener("click", (function () {
      let use_rank = cur_rank;
      return function (evt) {
        prevent_default(evt);
        do_show_less_parts(use_rank);
      };
    }) ());
  }
};

////////////////////////////////////////////////////////////////////////
// if got (possibly) image by any means

var setup_image_use = function (img_url, img_headers) {
  // it sometimes gets amounts of time to red a file,
  // thus putting in a timer, to prevent browser complaints;
  setTimeout(function () {
    setup_image_use_inner(img_url, img_headers);
  }, 0);
};

var setup_image_use_inner = function (img_url, img_headers) {
  debug_output("setup_image_use: " + img_url);

  set_image_notice("Preparing the image");
  show_image_notice(false);
  var to_show_more_less_rank = 10;

  if (bt_meta_exif && (!(bt_meta_exif.classList.contains(meta_section_not)))) {
    bt_meta_exif.classList.add(meta_section_not);
  }
  if (bt_meta_thumb && (!(bt_meta_thumb.classList.contains(meta_section_not)))) {
    bt_meta_thumb.classList.add(meta_section_not);
  }

  var thumb_items = document.getElementById("field-meta-thumb-items");
  if (thumb_items && (!(thumb_items.classList.contains(class_thumb_wo_items)))) {
    thumb_items.classList.add(class_thumb_wo_items);
    thumb_items.innerHTML = "";
  }
  var thumb_dims = document.getElementById("field-meta-thumb-dims");
  if (thumb_dims && (!(thumb_dims.classList.contains(class_thumb_wo_items)))) {
    thumb_dims.classList.add(class_thumb_wo_items);
  }
  var thumb_dims_width = document.getElementById("field-meta-thumb-width");
  var thumb_dims_height = document.getElementById("field-meta-thumb-height");
  if (thumb_dims_width) {
    thumb_dims_width.innerHTML = "";
  }
  if (thumb_dims_height) {
    thumb_dims_height.innerHTML = "";
  }
  if (can_thm_cvr_el) {
    can_thm_cvr_el.style.width = 0;
    can_thm_cvr_el.style.height = 0;
  }

  var thumb_subs = document.getElementById("field-meta-thumb-sub");

  var can_img = new Image();
  can_img.onerror = function () {
    set_image_notice("Could not use the image, possibly corrupted image.");
    show_image_notice(true);
    return;
  };
  can_img.onload = function () {
    image_file_start = image_file_start_loading;
    image_file_comment = image_file_comment_loading;
    png_xmp = png_xmp_loading;

    can_width = can_img.width;
    can_height = can_img.height;
    if ((!can_width) || (!can_height)) {
      set_image_notice("Empty image");
      show_image_notice(true);
      return;
    }

    if (img_size_width_el) {
      img_size_width_el.innerHTML = can_width;
    }
    if (img_size_height_el) {
      img_size_height_el.innerHTML = can_height;
    }

    img_aspect = can_width / can_height;

    can_ori_el.width = can_width;
    can_ori_el.height = can_height;

    can_res_levels_el.width = can_width;
    can_res_levels_el.height = can_height;

    can_res_slopes_el.width = can_width;
    can_res_slopes_el.height = can_height;

    can_res_ela_el.width = can_width;
    can_res_ela_el.height = can_height;

    can_res_pca_el.width = can_width;
    can_res_pca_el.height = can_height;

    can_aux_el.width = can_width;
    can_aux_el.height = can_height;

    can_ori_cvr_el.style.width = can_width;
    can_ori_cvr_el.style.height = can_height;
    can_res_cvr_el.style.width = can_width;
    can_res_cvr_el.style.height = can_height;

    can_ori_ctx = can_ori_el.getContext("2d");
    can_res_levels_ctx = can_res_levels_el.getContext("2d");
    can_res_slopes_ctx = can_res_slopes_el.getContext("2d");
    can_res_ela_ctx = can_res_ela_el.getContext("2d");
    can_res_pca_ctx = can_res_pca_el.getContext("2d");
    can_aux_ctx = can_aux_el.getContext("2d");

    can_ori_ctx.drawImage(can_img, 0, 0);

    let cur_vis_size = get_field_size();
    let cur_vis_width = cur_vis_size[0];
    let cur_vis_height = cur_vis_size[1];
    if (cur_vis_width && cur_vis_height) {
      if ((cur_vis_width >= can_width) && (cur_vis_height >= can_height)) {
        img_scaled = false;
      } else {
        img_scaled = true;
      }
      images_set_scaled();
    }

    try {
      EXIF.enableXmp();
      EXIF.getData(can_img, function() {
        var meta_data = EXIF.getAllTags(this);
        var exf_el = document.getElementById("field-meta-exif");

        var meta_part = "";
        var thumb_items_inner = "";
        var thumb_meta_el = document.getElementById("field-meta-thumb-img");
        for (var meta_type in meta_data) {
          if (!(meta_data.hasOwnProperty(meta_type))) {
            continue;
          }
          if (meta_type=="thumbnail") {
            var thumb_meta = meta_data["thumbnail"];
            for (var thumb_meta_type in thumb_meta) {
              if (thumb_meta_type == "blob") {
                continue;
              }
              thumb_items_inner += "<tr><td class=\"" + class_header_key + "\">" + safe_text_outer(thumb_meta_type) + "</td><td>";
              thumb_items_inner += safe_text_outer(JSON.stringify(thumb_meta[thumb_meta_type], null, " ")); + "</td></tr>";
            }
            if (("blob" in thumb_meta) && (thumb_meta["blob"])) {
              if (thumb_meta_el) {
                thumb_meta_el.onload = function () {
                  if (thumb_dims && (thumb_dims.classList.contains(class_thumb_wo_items))) {
                    thumb_dims.classList.remove(class_thumb_wo_items);
                  }
                  if (thumb_dims_width) {
                    thumb_dims_width.innerHTML = thumb_meta_el.width;
                  }
                  if (thumb_dims_height) {
                    thumb_dims_height.innerHTML = thumb_meta_el.height;
                  }
                  if (can_thm_cvr_el) {
                    can_thm_cvr_el.style.width = thumb_meta_el.width;
                    can_thm_cvr_el.style.height = thumb_meta_el.height;
                  }
                  if (thumb_subs) {
                    thumb_subs.style.marginTop = "" + thumb_meta_el.height + "px";
                  }
                };
                thumb_meta_el.src = URL.createObjectURL(meta_data["thumbnail"]["blob"]);
              }

              image_has_thumb = true;
              if (thumb_items_inner!="") {
                if (thumb_items && (thumb_items.classList.contains(class_thumb_wo_items))) {
                  thumb_items.classList.remove(class_thumb_wo_items);
                  thumb_items.innerHTML = thumb_items_inner;
                }
              }

            }
            continue;
          }
          let to_highlight_key = false;
          let to_highlight_other_key = false;
          let with_gps_key = false;
          let meta_type_lc = "";
          if (meta_type) {
            try {
              meta_type_lc = meta_type.toLowerCase();
            } catch (e) {
              meta_type_lc = "";
            }
          }
          if (meta_type_lc != "") {
            let types_to_highlight = ["datetime", "software"];
            for (let idx_hgh=types_to_highlight.length-1; idx_hgh>=0; idx_hgh--) {
              if (meta_type_lc.indexOf(types_to_highlight[idx_hgh]) > -1) {
                to_highlight_key = true;
                break;
              }
            }
            let types_to_highlight_other = ["copyright", "artist", "imagedescription"];
            for (let idx_hgh=types_to_highlight_other.length-1; idx_hgh>=0; idx_hgh--) {
              if (meta_type_lc.indexOf(types_to_highlight_other[idx_hgh]) > -1) {
                to_highlight_other_key = true;
                break;
              }
            }
            if (meta_type_lc.indexOf("gps") > -1) {
              with_gps_key = true;
            }
          }
          let one_title_part = "";
          let meta_type_use = "";
          if (typeof meta_type === "string") {
            meta_type_use = meta_type;
          } else {
            meta_type_use = JSON.stringify(meta_type, null, " ");
          }
          if (meta_type_use == "undefined") {
            meta_type_use = "unknown";
            meta_data[meta_type] = inner_trim_from_meta_array(meta_data[meta_type]);
            if (meta_data[meta_type]=="") {
              continue;
            }
          }
          image_has_exif = true;
          let uses_array_inner_string = false;
          if (meta_type_use.toLowerCase && (meta_type_use.toLowerCase() in {"usercomment":true, "makernote":true, "unknown":true})) {
            if (typeof meta_data[meta_type] === "object") {
              if (inner_empty_from_meta_array(meta_data[meta_type])) {
                meta_data[meta_type] = "";
              }
            }
            if (typeof meta_data[meta_type] === "object") {
              try {
                let one_unistr_inner = inner_unicode_from_meta_array(meta_data[meta_type]);
                if (one_unistr_inner == "") {
                  one_unistr_inner = inner_ascii_from_meta_array(meta_data[meta_type]);
                }
                if (one_unistr_inner != "") {
                  meta_data[meta_type] = one_unistr_inner;
                  uses_array_inner_string = true;
                  to_highlight_other_key = true;
                }
              } catch (e) {
                uses_array_inner_string = false;
              }
              if (!uses_array_inner_string) {
                let one_title_inner = inner_string_from_meta_array(meta_data[meta_type]);
                if (one_title_inner != "") {
                  one_title_part = " title=\"" + one_title_inner + "\"";
                }
              }
            }
          }

          let key_class_part = " class=\"" + class_header_key;
          let val_class_part = " class=\"";
          if (to_highlight_key) {
            key_class_part += " " + class_header_highlight;
            val_class_part += " " + class_header_highlight;
          }
          if (to_highlight_other_key) {
            key_class_part += " " + class_header_highlight_other;
            val_class_part += " " + class_header_highlight_other;
          }
          if (with_gps_key) {
            key_class_part += " " + class_header_gps;
          }
          key_class_part += "\"";
          val_class_part += "\"";

          if (typeof meta_data[meta_type] === "string") {
            meta_data[meta_type] = meta_data[meta_type].replace(/\0+$/g, "");
            let start_meta_value = meta_data[meta_type];
            let fixed_meta_value = "";
            try {
              fixed_meta_value = decodeURIComponent(escape(start_meta_value));
            } catch (e) {
              fixed_meta_value = "";
            }
            if ((fixed_meta_value != "") && (fixed_meta_value != start_meta_value)) {
              meta_data[meta_type] = fixed_meta_value;
            }
          }
          meta_part += "<tr><td" + key_class_part + ">" + safe_text_outer(meta_type_use);
          meta_part += "</td><td" + val_class_part + one_title_part + ">";
          if (meta_type_use.toLowerCase && (meta_type_use.toLowerCase() == "makernote")) {
            let makernote_val = (uses_array_inner_string) ? meta_data[meta_type] : safe_text_outer(JSON.stringify(meta_data[meta_type], null, " "));
            let makernote_maxlen = 400;
            if (makernote_val.length <= (makernote_maxlen+100)) {
              meta_part += makernote_val;
            } else {
              to_show_more_less_rank += 1;
              let check_idx = -1;
              while (check_idx < 20) {
                check_idx += 1;
                if ((makernote_val.charAt(makernote_maxlen + check_idx)) in {",":true, ";":true, " ":true}) {
                  break;
                }
              }
              makernote_maxlen += (check_idx + 1);
              let class_part_less_one = class_part_less + "-" + to_show_more_less_rank;
              let class_part_more_one = class_part_more + "-" + to_show_more_less_rank;
              meta_part += makernote_val.substring(0, makernote_maxlen);
              meta_part += "<span class=\"" + class_part_less_one +"\">... </span>";
              meta_part += "<span id=\"id-" + class_show_more + "-" + to_show_more_less_rank + "\" class=\"" + class_part_less_one + " " + class_show_more + " bt-show-more-less without-selection\" tabIndex=\"" + to_show_more_less_rank + "\">show&nbsp;more</span>";
              meta_part += "<span class=\"" + class_part_more_one + " " + class_el_hidden + "\">";
              meta_part += makernote_val.substring(makernote_maxlen) + " </span>";
              meta_part += "<span id=\"id-" + class_show_less + "-" + to_show_more_less_rank + "\" class=\"" + class_part_more_one + " " + class_el_hidden + " " + class_show_less + " bt-show-more-less without-selection\" tabIndex=\"" + to_show_more_less_rank + "\">show&nbsp;less</span>";
            }
          } else {
            if (!uses_array_inner_string) {
              meta_part += safe_text_outer(JSON.stringify(meta_data[meta_type], null, " ")).replace(/\\r\\n"$|\\r"$|\\n"$/, "\"").split(/\\r\\n|\\r|\\n/g).join("<br>");
            } else {
              meta_part += meta_data[meta_type];
            }
          }
          meta_part += "</td></tr>\n";
        }

        let meta_data_iptc = EXIF.getAllIptcTags(this);
        let notified_on_iptc = false;
        for (var meta_type in meta_data_iptc) {
          if (!(meta_data_iptc.hasOwnProperty(meta_type))) {
            continue;
          }
          if (!notified_on_iptc) {
            notified_on_iptc = true;
            if (image_has_exif) {
              meta_part += "<tr><td class=\"" + class_header_key + "\">" + "&nbsp;";
              meta_part += "</td><td>&nbsp;</td></tr>\n";
            }
            meta_part += "<tr><td class=\"" + class_header_key + " " +  class_header_highlight + "\">" + "IPTC tags:";
            meta_part += "</td><td>&nbsp;</td></tr>\n";
          }

          image_has_exif = true;
          let key_class_part = " class=\"" + class_header_key + "\"";

          let val_class_part = "";
          let one_title_part = "";

          let meta_type_use = "";
          if (typeof meta_type === "string") {
            meta_type_use = meta_type;
          } else {
            meta_type_use = JSON.stringify(meta_type, null, " ");
          }
          if (typeof meta_data_iptc[meta_type] === "string") {
            meta_data_iptc[meta_type] = meta_data_iptc[meta_type].replace(/\0+$/g, "");
            let start_meta_value = meta_data_iptc[meta_type];
            let fixed_meta_value = "";
            try {
              fixed_meta_value = decodeURIComponent(escape(start_meta_value));
            } catch (e) {
              fixed_meta_value = "";
            }
            if ((fixed_meta_value != "") && (fixed_meta_value != start_meta_value)) {
              meta_data_iptc[meta_type] = fixed_meta_value;
            }
          }
          meta_part += "<tr><td" + key_class_part + ">" + safe_text_outer(meta_type_use);
          meta_part += "</td><td" + val_class_part + one_title_part + ">";
          meta_part += safe_text_outer(JSON.stringify(meta_data_iptc[meta_type], null, " ")).replace(/\\r\\n"$|\\r"$|\\n"$/, "\"").split(/\\r\\n|\\r|\\n/g).join("<br>") + "</td></tr>\n";
        }

        let xmpdata_parts = {};
        if (can_img.xmpdata) {
          xmpdata_parts = can_img.xmpdata;
        }
        let xmpdata_some = false;
        for (xmp_test_key in xmpdata_parts) {
          if (!(xmpdata_parts.hasOwnProperty(xmp_test_key))) {
            continue;
          }
          xmpdata_some = true;
          break;
        }
        if ((!xmpdata_some) && png_xmp) {
          xmpdata_parts = png_xmp;
        }

        if (xmpdata_parts["x:xmpmeta"]) {
          xmpdata_parts = xmpdata_parts["x:xmpmeta"];
        }
        if (xmpdata_parts["rdf:RDF"]) {
          xmpdata_parts = xmpdata_parts["rdf:RDF"];
        }
        if (xmpdata_parts["rdf:Description"]) {
          xmpdata_parts = xmpdata_parts["rdf:Description"];
        }
        if (xmpdata_parts) {
          if (typeof xmpdata_parts[0] === "undefined") {
            xmpdata_parts = [xmpdata_parts];
          }
        }

        if (xmpdata_parts) {
          let notified_on_xmp = false;
          let xmpdata_parts_len = xmpdata_parts.length;
          for (let xmpdata_parts_key in xmpdata_parts) {
            if (!(xmpdata_parts.hasOwnProperty(xmpdata_parts_key))) {
              continue;
            }
            let xmpdata = xmpdata_parts[xmpdata_parts_key];
            for (let meta_type in xmpdata) {
              if (!(xmpdata.hasOwnProperty(meta_type))) {
                continue;
              }
              if (!notified_on_xmp) {
                notified_on_xmp = true;
                if (image_has_exif) {
                  meta_part += "<tr><td class=\"" + class_header_key + "\">" + "&nbsp;";
                  meta_part += "</td><td>&nbsp;</td></tr>\n";
                }
                meta_part += "<tr><td class=\"" + class_header_key + " " +  class_header_highlight + "\">" + "XMP data:";
                meta_part += "</td><td>&nbsp;</td></tr>\n";
              }

              image_has_exif = true;

              let meta_type_use = "";
              if (typeof meta_type === "string") {
                meta_type_use = meta_type;
              } else {
                meta_type_use = JSON.stringify(meta_type, null, " ");
              }
              meta_type_use = safe_text_outer(meta_type_use);

              let with_gps_key = false;
              let type_is_digest = false;
              let to_highlight_key = false;
              if (typeof meta_type_use === "string") {
                if (meta_type_use.length > 20) {
                  meta_type_use = meta_type_use.replace(/\:/g, ":<br>");
                } else {
                  meta_type_use = meta_type_use.replace(/\:/g, ":<wbr>");
                }
                if (meta_type_use.toLowerCase().indexOf("digest") > -1) {
                  type_is_digest = true;
                }
                if (meta_type_use.toLowerCase().indexOf("date") > -1) {
                  to_highlight_key = true;
                }
                if (meta_type_use.toLowerCase().indexOf("gps") > -1) {
                  with_gps_key = true;
                }
              }

              let key_class_part = " class=\"" + class_header_key;
              let val_class_part = " class=\"";
              if (with_gps_key) {
                key_class_part += " " + class_header_gps;
              }
              key_class_part += "\"";
              val_class_part += "\"";

              let meta_value = xmpdata[meta_type];
              if (typeof meta_value === "object") {
                let has_text_prop = false;
                let has_text_prop_only = true;
                for (let test_obj_key in meta_value) {
                  if (test_obj_key == "#text") {
                    has_text_prop = true;
                  } else {
                    has_text_prop_only = false;
                    break;
                  }
                }
                if (has_text_prop && has_text_prop_only) {
                  meta_value = meta_value["#text"];
                }
              }
              meta_value = JSON.stringify(meta_value, null, 4);
              let fixed_meta_value = "";
              try {
                fixed_meta_value = decodeURIComponent(escape(meta_value));
              } catch (e) {
                fixed_meta_value = "";
              }
              if (fixed_meta_value != "") {
                meta_value = fixed_meta_value;
              }
              let line_start_place = "";
              for (idx_lsp=0; idx_lsp<6; idx_lsp++) {
                line_start_place += "&nbsp;";
              }
              meta_value = safe_text_outer(meta_value).replace(/"stEvt:when"/g, "\"<b>stEvt:when</b>\"")
                .replace(/\:ModifyDate/g, ":<b>ModifyDate</b>")
                .replace(/\:CreateDate/g, ":<b>CreateDate</b>")
                .replace(/\:MetadataDate/g, ":<b>MetadataDate</b>")
                .replace(/\:DateCreated/g, "\:<b>DateCreated</b>");
              meta_part += "<tr><td" + key_class_part + ">" + meta_type_use;
              meta_part += "</td><td" + val_class_part + ">";
              meta_part += "<table border=\"1\" class=\"meta-xmp-value-holder-outer\"><tr><td class=\"meta-xmp-value-holder-inner\">";
              if (type_is_digest) {
                meta_part += "<span class=\"meta-xmp-digest-value\">" + meta_value.replace(/\r\n|\r|\n/g, "<br>\n").replace(/;/g, ";<br>").replace(/\,/g, ",<wbr>")
                  .replace(/"\#text"/g, line_start_place + "\"#text\"") + "</span>";
              } else {
                meta_part += "<pre class=\"meta-xmp-pre-value\">" + (to_highlight_key ? "<b>": "");
                meta_part += meta_value.replace(/\\r\\n|\\r|\\n/g, "\n").replace(/;/g, ";<wbr>").replace(/\,/g, ",<wbr>");
                meta_part += (to_highlight_key ? "</b>": "") + "</pre>";
              }
              meta_part += "</td></tr></table>";
              meta_part += "</td></tr>\n";
            }
          }
        }

        if (exf_el && (meta_part != "")) {
          exf_el.innerHTML = meta_part;
        }

        disp_part_meta = "file";
        if (image_has_thumb) {
          disp_part_meta = "thumb";
          if (bt_meta_thumb && (bt_meta_thumb.classList.contains(meta_section_not))) {
            bt_meta_thumb.classList.remove(meta_section_not);
          }
        }
        if (image_has_exif) {
          disp_part_meta = "exif";
          if (bt_meta_exif && (bt_meta_exif.classList.contains(meta_section_not))) {
            bt_meta_exif.classList.remove(meta_section_not);
          }
        }

        hide_image_notice();
        set_image_is_loaded(true);
        show_image_base(img_headers);
        main_use_part("meta");

        setTimeout(function () {
          setup_show_more_less_parts();
        }, 10);
      });
    }
    catch (e) {
      disable_exif();
      disable_thumb();
      set_image_notice("Can not read image metadata");
      show_image_notice(true);
      hide_image_notice(default_notice_timeout);
      set_image_is_loaded(true);
      show_image_base(img_headers);
      main_use_part("meta");
      setTimeout(function () {
        setup_show_more_less_parts();
      }, 10);
    }
  };

  show_image_notice(true);
  set_image_is_loaded(false);
  zero_ini_values();
  can_img.src = img_url;
};

////////////////////////////////////////////////////////////////////////
// after has asked for image by URL

var utilize_image_from_url = function (ans_state) {
  debug_output("utilize_image_from_url: " + ans_state);

  image_file_start_loading = "";
  image_file_comment_loading = "";
  png_xmp_loading = {};

  img_request_runs = false;
  if (!img_request) {
    return;
  }
  if (!ans_state) {
    img_request = null;
    return;
  }

  var image_type = "image/jpeg";

  var resp_headers_bare = img_request.getAllResponseHeaders();
  if (resp_headers_bare) {
    resp_headers_bare = resp_headers_bare.split("\n");
  } else {
    resp_headers_bare = [];
  }
  var resp_headers_len = resp_headers_bare.length;
  var resp_headers = [];
  for (let one_resp_part_idx = 0; one_resp_part_idx < resp_headers_len; one_resp_part_idx++) {
    let one_resp_part = resp_headers_bare[one_resp_part_idx];
    let one_resp_header_key = "";
    let one_resp_header_value = "";
    let one_resp_header_key_end = one_resp_part.indexOf(":");
    if (one_resp_header_key_end != -1) {
      one_resp_header_key = one_resp_part.substring(0, one_resp_header_key_end);
      one_resp_header_value = one_resp_part.substring(one_resp_header_key_end+1);
    } else {
      one_resp_header_key = one_resp_part;
    }
    if (one_resp_header_key.toLowerCase() == "content-type") {
      image_type = one_resp_header_value;
    }
    resp_headers.push([one_resp_header_key, one_resp_header_value]);
  }

  var img_data = null;
  if ("mozResponseArrayBuffer" in img_request) {
    img_data = img_request.mozResponseArrayBuffer;
  }
  if ((!img_data) && img_request.hasOwnProperty("responseType") && img_request.responseType == "arraybuffer") {
    img_data = img_request.response;
  }
  if ((!img_data) && ("response" in img_request)) {
    img_data = img_request.response;
  }
  img_request = null;
  if (!img_data) {
    set_image_notice("No image data in the downloaded file");
    show_image_notice(true);
    if (image_is_loaded) {
      hide_image_notice(default_notice_timeout);
    }
    return;
  }

  try {
    let img_data_array = new Uint8Array(img_data);
    let img_data_len_test = img_data_array.length;
    let line_length = 80;
    let first_line_end = line_length;
    if (first_line_end > img_data_len_test) {
      first_line_end = img_data_len_test;
    }

    // an arbitrary length for comment testing
    let img_data_len_test_max = 75537;
    if (img_data_len_test > img_data_len_test_max) {
      img_data_len_test = img_data_len_test_max;
    }

    let test_image_start = "";
    for (let idx=0; idx < img_data_len_test; idx++) {
      test_image_start += String.fromCharCode(img_data_array[idx]);
    }
    image_file_comment_loading = image_file_extract_comment_pngxmp(test_image_start, image_type);

    for (let ind_str=0; ind_str < first_line_end; ind_str++) {
      image_file_start_loading += "&#" + img_data_array[ind_str] + ";";
    }
    if (img_data_len_test > line_length) {
      image_file_start_loading += "<br>";
      let second_line_end = 2*line_length;
      if (second_line_end > img_data_len_test) {
        second_line_end = img_data_len_test;
      }
      for (let ind_str=first_line_end; ind_str<second_line_end; ind_str++) {
        image_file_start_loading += "&#" + img_data_array[ind_str] + ";";
      }
    }

    img_data_array = null;
  } catch (e) {
    image_file_start_loading = "could not read file content";
    image_file_comment_loading = "";
    png_xmp_loading = {};
  }

  var img_blob = new Blob([img_data], {type: "application/octet-binary"});
  var img_url = URL.createObjectURL(img_blob);

  setup_image_use(img_url, resp_headers);
};

////////////////////////////////////////////////////////////////////////
// asking for image by URL

var stop_previous_image_loading = function () {
  debug_output("stop_previous_image_loading");

  if (img_request_runs) {
    if (img_request) {
      try {img_request.abort();}
      catch (e) {no_op_function();}
      img_request = null;
    }
    img_request_runs = false;
  }
};

var load_new_image_from_url = function (use_url) {
  debug_output("load_new_image_from_url: " + use_url);

  history.replaceState(undefined, undefined, "#" + use_url);
  // would need to add listening on URL changes
  // if dealt with pushState() instead of replaceState();

  load_image_from_url();
};

var load_image_from_url = function () {
  debug_output("load_image_from_url");

  var page_url = window.location.href.toString();
  var hash_pos = page_url.indexOf("#");
  if (hash_pos == -1) {
    if (!image_is_loaded) {
      set_image_notice("No image loaded");
      show_image_notice(false);
      if (image_is_loaded) {
        hide_image_notice(default_notice_timeout);
      }
    }
    return;
  }
  var img_url = page_url.substring(hash_pos+1);
  if (img_url.substring(0, 4)!="http") {
    if (!image_is_loaded) {
      set_image_notice("No image loaded");
      show_image_notice(false);
      if (image_is_loaded) {
        hide_image_notice(default_notice_timeout);
      }
    }
    return;
  }

  set_image_notice("Preparing to load image");
  show_image_notice(true);

  stop_previous_image_loading();

  img_request = new XMLHttpRequest();
  if (!img_request) {
    set_image_notice("Can not load image");
    show_image_notice(true);
    if (image_is_loaded) {
      hide_image_notice(default_notice_timeout);
    }
    return;
  }
  img_request.open('GET', img_url, true);
  img_request.responseType = "arraybuffer";
  img_request.setRequestHeader("Accept", "*/*");

  img_request.onreadystatechange = function () {
    if (!img_request) {
      set_image_notice("Could not load image");
      show_image_notice(true);
      if (image_is_loaded) {
        hide_image_notice(default_notice_timeout);
      }
      return;
    }
    var req_done = img_request.DONE || 4;
    if (img_request.readyState === req_done) {
      if (img_request.status!=200) {
        set_image_notice("Image was not downloaded");
        show_image_notice(true);
        if (image_is_loaded) {
          hide_image_notice(default_notice_timeout);
        }
        utilize_image_from_url(false);
        return;
      }
      set_image_notice("Image got downloaded");
      show_image_notice(false);

      last_loading_path = img_url;
      last_loading_local = false;
      utilize_image_from_url(true);
    }
  };

  set_image_notice("Loading image");
  show_image_notice(true);

  img_request.send(null);
  img_request_runs = true;
};

////////////////////////////////////////////////////////////////////////
// handling local image file

var handle_local_image_file = function (img_file, dragged) {
  debug_output("handle_local_image_file");

  stop_previous_image_loading();

  var fl_el = document.getElementById("in-load-file");

  if ((!(img_file.type)) || (!(img_file.type.match(/image.*/)))) {
    image_loaded_none("Image type not recognized");
    return;
  }
  var reader = new FileReader();
  reader.onerror = function (evt) {
    if (!dragged) {
      fl_el.value = "";
    }
    image_loaded_none("Can not read the image, possibly without read permissions.");
    return;
  };
  reader.onload = (function () {
    if (!dragged) {
      fl_el.value = "";
    }
    let file_last_modified = "";
    let file_content_type = "";
    let file_content_length = "";
    if (img_file.lastModified) {
      let img_mod_date = new Date(img_file.lastModified);
      file_last_modified = img_mod_date.toUTCString();
    }
    if (img_file.type) {
      file_content_type = img_file.type;
    }
    if (img_file.size) {
      file_content_length = img_file.size;
    }
    return function (evt) {
      if (evt && evt.target && evt.target.result) {
        let res_data = evt.target.result;
        let comma_pos = res_data.indexOf(",");
        if (comma_pos>-1) {
          let guessed_type = guess_image_type(res_data.substring(comma_pos+1, comma_pos+9), true);
          if (guessed_type != "") {
            file_content_type = guessed_type;
          }
          image_file_start_loading = "";
          image_file_comment_loading = "";
          png_xmp_loading = {};
          try {
            // an arbitrary length (just divisible by 4, so that it can get decoded from base64) for comment testing
            let img_start = window.atob(res_data.substring(comma_pos+1,comma_pos+100717));
            let img_start_len = img_start.length;
            image_file_comment_loading = image_file_extract_comment_pngxmp(img_start, file_content_type);
            let line_length = 80;
            let first_line_end = line_length;
            if (first_line_end > img_start_len) {
              first_line_end = img_start_len;
            }
            for (let ind_str=0; ind_str<first_line_end; ind_str++) {
              image_file_start_loading += "&#" + img_start.charCodeAt(ind_str) + ";";
            }
            if (img_start_len > line_length) {
              image_file_start_loading += "<br>";
              let second_line_end = 2*line_length;
              if (second_line_end > img_start_len) {
                second_line_end = img_start_len;
              }
              for (let ind_str=first_line_end; ind_str<second_line_end; ind_str++) {
                image_file_start_loading += "&#" + img_start.charCodeAt(ind_str) + ";";
              }
            }
          } catch (e) {
            image_file_start_loading = "could not read file content";
            image_file_comment_loading = "";
            png_xmp_loading = {};
          }
        }
        var file_headers = [];
        if (file_last_modified != "") {
          file_headers.push(["last-modified", file_last_modified]);
        }
        if (file_content_type != "") {
          file_headers.push(["content-type", file_content_type]);
        }
        if (file_content_length != "") {
          file_headers.push(["content-length", file_content_length]);
        }
        setup_image_use(res_data, file_headers);
      } else {
        image_loaded_none();
      }
    };
  }) ();
  if (img_file.name) {
    last_loading_path = img_file.name;
  } else {
    last_loading_path = "";
  }
  last_loading_local = true;
  reader.readAsDataURL(img_file);
};

