////////////////////////////////////////////////////////////////////////
// auxiliary

var no_op_function = function () {
};

var debug_output = function (data) {
  if (under_debug) {
    let to_output  = data;
    if ((typeof to_output === "string") && (to_output.length > 200)) {
      to_output = to_output.substring(0, 200) + "...";
    }
    console.log(to_output);
  }
};

var prevent_default = function (evt) {
  evt = evt || window.event;
  if (typeof evt !== "undefined") {
    if (evt.preventDefault) {evt.preventDefault();}
    if (evt.stopPropagation) {evt.stopPropagation();}
  }
};

var safe_text_outer = function (input) {
  if (typeof input !== "string") {
    // this should only happen for numbers
    return input;
  }
  return input.replace(/\&/g, "&amp;").replace(/\</g, "&lt;").replace(/\>/g, "&gt;");
};

var safe_text_inner = function (input) {
  if (typeof input !== "string") {
    // this should only happen for numbers
    return input;
  }
  return input.replace(/\&/g, "&amp;").replace(/\</g, "&lt;").replace(/\>/g, "&gt;").replace(/\"/g, "&quot;").replace(/\'/g, "&#39;").replace(/\\/g, "&#92;");
};

var get_field_size = function () {
  var vis_el = document.getElementById("fields-holder");

  if (!vis_el) {
    return [100,100];
  }
  return [vis_el.clientWidth, vis_el.clientHeight];
};

////////////////////////////////////////////////////////////////////////
// setting the initial values

var zero_ini_values = function () {
  debug_output("zero_ini_values");
  // this is called during initial setup and at loading of a (new) image;

  gray_levels_val = false;
  spread_levels_val = true;
  center_levels_val = 0.5;
  contrast_levels_val = 10;

  form_slopes = "any";
  lumin_slopes = false;
  step_slopes = 1;
  light_slopes = 20;

  pca_eigen_rank = 1;
  pca_eigen_mode = 0;
  light_pca = 10;
  light_pca_display = "2.0";
  light_pca_use = "2.0";

  blur_ela = 0.25;
  light_ela = 20;
  boost_ela = false;

  set_levels_knobs();
  set_levels_knobs_center();
  set_levels_knobs_contrast();
  set_levels_display_center();
  set_levels_display_contrast();
  clean_levels_values();

  set_pca_knobs();
  set_pca_display_light();
  set_pca_display_knobs();
  clean_pca_values();

  set_slopes_knobs();
  set_slopes_knobs_light();
  set_slopes_display_light();
  clean_slopes_values();

  set_ela_knobs();
  set_ela_display_blur();
  set_ela_display_light();
  set_ela_display_knobs();
  clean_ela_values();
  set_ela_computing(false);

  images_show_stage(img_result_form);
};

////////////////////////////////////////////////////////////////////////
// setting handlers for events within UI

var setup_page_elements = function () {
  debug_output("setup_page_elements");

  file_notice_hld_el = document.getElementById("field-meta-file-notice-holder");
  file_notice_el = document.getElementById("field-meta-file-notice");

  var bts_scaled = {
    "bt-scaled-on": true,
    "bt-scaled-off": true,
    "bt-scaled-label": true
  };
  for (var el_scl_id in bts_scaled) {
    if (!(bts_scaled.hasOwnProperty(el_scl_id))) {
      continue;
    }
    let bt_el = document.getElementById(el_scl_id);
    if (bt_el) {
      bt_el.onclick = (function () {
        return function (evt) {
          images_change_scaled(evt);
        };
      })();
    }
  }

  var bts_parts = {
    "bt-main-meta": "meta",
    "bt-main-levels": "levels",
    "bt-main-slopes": "slopes",
    "bt-main-ela": "ela",
    "bt-main-pca": "pca",
    "bt-main-about": "about"
  };
  for (var el_par_id in bts_parts) {
    if (!(bts_parts.hasOwnProperty(el_par_id))) {
      continue;
    }
    let bt_el = document.getElementById(el_par_id);
    if (bt_el) {
      bt_el.onclick = (function () {
        var use_part = bts_parts[el_par_id];
        return function (evt) {
          main_use_part_outer(evt, use_part);
        };
      })();
    }
  }

  var bts_images = {
    "bt-images-ini": "ini",
    "bt-images-mix": "mix",
    "bt-images-fin": img_result_form
  };
  for (var el_img_id in bts_images) {
    if (!(bts_images.hasOwnProperty(el_img_id))) {
      continue;
    }
    let bt_el = document.getElementById(el_img_id);
    if (bt_el) {
      bt_el.onclick = (function () {
        var stage_show = bts_images[el_img_id];
        return function (evt) {
          images_show_stage_outer(evt, stage_show);
        };
      })();
    }
  }

  var levels_grayscaled_el = document.getElementById("bt-levels-alt-gray");
  if (levels_grayscaled_el) {
    levels_grayscaled_el.addEventListener("click", function (evt) {
      switch_levels_gray_outer(evt);
    });
    levels_grayscaled_el.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        switch_levels_gray_outer(evt);
      }
    });
  }

  var levels_spread_el = document.getElementById("bt-levels-alt-spread");
  if (levels_spread_el) {
    levels_spread_el.addEventListener("click", function (evt) {
      switch_levels_spread_outer(evt);
    });
    levels_spread_el.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        switch_levels_spread_outer(evt);
      }
    });
  }

  var levels_contrast_el = document.getElementById("rn-levels-contrast");
  if (levels_contrast_el) {
    levels_contrast_el.addEventListener("input", function (evt) {
      if (levels_contrast_el) {
        contrast_levels_val = levels_contrast_el.value;
        set_levels_display_contrast();
      }
      update_filters_outer(evt);
    });
  }

  var levels_cent_el = document.getElementById("rn-levels-center");
  if (levels_cent_el) {
    levels_cent_el.addEventListener("input", function (evt) {
      if (levels_cent_el) {
        center_levels_val = levels_cent_el.value;
        set_levels_display_center();
      }
      update_filters_outer(evt);
    });
  }

  if (bt_meta_exif) {
    bt_meta_exif.addEventListener("click", function (evt) {
      prevent_default(evt);
      if (image_has_exif) {
        disp_part_meta = "exif";
        make_current_meta(true);
      } else {
        if (!image_is_loaded) {
          blink_load_button();
        }
      }
    });
  }
  if (bt_meta_thumb) {
    bt_meta_thumb.addEventListener("click", function (evt) {
      prevent_default(evt);
      if (image_has_thumb) {
        disp_part_meta = "thumb";
        make_current_meta(true);
      } else {
        if (!image_is_loaded) {
          blink_load_button();
        }
      }
    });
  }
  if (bt_meta_file) {
    bt_meta_file.addEventListener("click", function (evt) {
      prevent_default(evt);
      disp_part_meta = "file";
      make_current_meta(true);
    });
  }

  let bt_about_forensics = document.getElementById("ctrl-about-forensics");
  if (bt_about_forensics) {
    bt_about_forensics.addEventListener("click", function (evt) {
      prevent_default(evt);
      disp_part_about = "forensics";
      make_current_about();
    });
  }
  let bt_about_meta = document.getElementById("ctrl-about-meta");
  if (bt_about_meta) {
    bt_about_meta.addEventListener("click", function (evt) {
      prevent_default(evt);
      disp_part_about = "meta";
      make_current_about();
    });
  }
  let bt_about_levels = document.getElementById("ctrl-about-levels");
  if (bt_about_levels) {
    bt_about_levels.addEventListener("click", function (evt) {
      prevent_default(evt);
      disp_part_about = "levels";
      make_current_about();
    });
  }
  let bt_about_slopes = document.getElementById("ctrl-about-slopes");
  if (bt_about_slopes) {
    bt_about_slopes.addEventListener("click", function (evt) {
      prevent_default(evt);
      disp_part_about = "slopes";
      make_current_about();
    });
  }
  let bt_about_ela = document.getElementById("ctrl-about-ela");
  if (bt_about_ela) {
    bt_about_ela.addEventListener("click", function (evt) {
      prevent_default(evt);
      disp_part_about = "ela";
      make_current_about();
    });
  }
  let bt_about_pca = document.getElementById("ctrl-about-pca");
  if (bt_about_pca) {
    bt_about_pca.addEventListener("click", function (evt) {
      prevent_default(evt);
      disp_part_about = "pca";
      make_current_about();
    });
  }
  let bt_about_image = document.getElementById("ctrl-about-image");
  if (bt_about_image) {
    bt_about_image.addEventListener("click", function (evt) {
      prevent_default(evt);
      disp_part_about = "image";
      make_current_about();
    });
  }

  let in_load_url = document.getElementById("in-load-url");
  if (in_load_url) {
    in_load_url.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((evt.which != 13) && (evt.keyCode != 13)) {
        return true;
      }
      if (!(in_load_url.value.match(/^http(s)?\:\/\/.+/i))) {
        return true;
      }
      prevent_default(evt);
      load_new_image_from_url(in_load_url.value);
      return false;
    });
  }

  let bt_slopes_form_any = document.getElementById("bt-slopes-form-any");
  if (bt_slopes_form_any) {
    bt_slopes_form_any.addEventListener("click", function (evt) {
      form_slopes = "any";
      make_current_slopes_outer(evt);
    });
    bt_slopes_form_any.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        form_slopes = "any";
        make_current_slopes_outer(evt);
      }
    });
  }
  let bt_slopes_form_steady = document.getElementById("bt-slopes-form-steady");
  if (bt_slopes_form_steady) {
    bt_slopes_form_steady.addEventListener("click", function (evt) {
      form_slopes = "steady";
      make_current_slopes_outer(evt);
    });
    bt_slopes_form_steady.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        form_slopes = "steady";
        make_current_slopes_outer(evt);
      }
    });
  }
  let bt_slopes_form_grainy = document.getElementById("bt-slopes-form-grainy");
  if (bt_slopes_form_grainy) {
    bt_slopes_form_grainy.addEventListener("click", function (evt) {
      form_slopes = "grainy";
      make_current_slopes_outer(evt);
    });
    bt_slopes_form_grainy.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        form_slopes = "grainy";
        make_current_slopes_outer(evt);
      }
    });
  }
  let bt_slopes_form_blur = document.getElementById("bt-slopes-form-blur");
  if (bt_slopes_form_blur) {
    bt_slopes_form_blur.addEventListener("click", function (evt) {
      form_slopes = "blur";
      make_current_slopes_outer(evt);
    });
    bt_slopes_form_blur.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        form_slopes = "blur";
        make_current_slopes_outer(evt);
      }
    });
  }

  let bt_slopes_modif_lumin = document.getElementById("bt-slopes-modif-lumin");
  if (bt_slopes_modif_lumin) {
    bt_slopes_modif_lumin.addEventListener("click", function (evt) {
      lumin_slopes = (!lumin_slopes);
      make_current_slopes_outer(evt);
    });
    bt_slopes_modif_lumin.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        lumin_slopes = (!lumin_slopes);
        make_current_slopes_outer(evt);
      }
    });
  }

  let bt_slopes_step_double = document.getElementById("bt-slopes-modif-double");
  if (bt_slopes_step_double) {
    bt_slopes_step_double.addEventListener("click", function (evt) {
      step_slopes = 3 - step_slopes;
      make_current_slopes_outer(evt);
    });
    bt_slopes_step_double.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        step_slopes = 3 - step_slopes;
        make_current_slopes_outer(evt);
      }
    });
  }

  slopes_light_el = document.getElementById("rn-slopes-light");
  if (slopes_light_el) {
    slopes_light_el.addEventListener("input", function (evt) {
      light_slopes = slopes_light_el.value;
      set_slopes_display_light();
      update_filters_outer(evt);
    });
  }

  ela_light_el = document.getElementById("rn-ela-light");
  if (ela_light_el) {
    ela_light_el.addEventListener("input", function (evt) {
      light_ela = ela_light_el.value;
      set_ela_display_light();
      update_filters_outer(evt);
    });
  }

  ela_blur_el = document.getElementById("rn-ela-blur");
  if (ela_blur_el) {
    ela_blur_el.addEventListener("input", function (evt) {
      blur_ela = ela_blur_el.value;
      set_ela_display_blur();
      make_current_ela_outer(evt);
    });
    ela_blur_el.addEventListener("keydown", function (evt) {
      make_ela_delay = 500;
      let curr_date = new Date();
      make_ela_delay_set = curr_date.getTime();
    });
    ela_blur_el.addEventListener("keyup", function (evt) {
      make_ela_delay = 50;
      blur_ela = ela_blur_el.value;
      set_ela_display_blur();
      make_current_ela_outer(evt);
    });
    ela_blur_el.addEventListener("wheel", function (evt) {
      make_ela_delay = 500;
      let curr_date = new Date();
      make_ela_delay_set = curr_date.getTime();
    });
    ela_blur_el.addEventListener("mousemove", function (evt) {
      make_ela_delay = 500;
      let curr_date = new Date();
      make_ela_delay_set = curr_date.getTime();
    });
    ela_blur_el.addEventListener("mousedown", function (evt) {
      make_ela_delay = 200;
      let curr_date = new Date();
      make_ela_delay_set = curr_date.getTime();
    });
    ela_blur_el.addEventListener("mouseup", function (evt) {
      make_ela_delay = 50;
      let curr_date = new Date();
      make_ela_delay_set = curr_date.getTime();
    });
    ela_blur_el.addEventListener("click", function (evt) {
      make_ela_delay = 50;
      let curr_date = new Date();
      make_ela_delay_set = curr_date.getTime();
    });
  }

  let ela_modif_boosted = document.getElementById("bt-ela-modif-boosted");
  if (ela_modif_boosted) {
    ela_modif_boosted.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        if (ela_inner_working) {
          prevent_default(evt);
          return;
        }
        boost_ela = !boost_ela;
        set_ela_display_knobs();
        make_current_ela_outer(evt);
      }
    });
    ela_modif_boosted.addEventListener("click", function (evt) {
      if (ela_inner_working) {
        prevent_default(evt);
        return;
      }
      boost_ela = !boost_ela;
      set_ela_display_knobs();
      make_current_ela_outer(evt);
    });
  }

  let pca_component_first = document.getElementById("bt-pca-component-first");
  if (pca_component_first) {
    pca_component_first.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        pca_eigen_rank = 0;
        set_pca_display_knobs();
        make_current_pca_outer(evt);
      }
    });
    pca_component_first.addEventListener("click", function (evt) {
      pca_eigen_rank = 0;
      set_pca_display_knobs();
      make_current_pca_outer(evt);
    });
  }

  let pca_component_second = document.getElementById("bt-pca-component-second");
  if (pca_component_second) {
    pca_component_second.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        pca_eigen_rank = 1;
        set_pca_display_knobs();
        make_current_pca_outer(evt);
      }
    });
    pca_component_second.addEventListener("click", function (evt) {
      pca_eigen_rank = 1;
      set_pca_display_knobs();
      make_current_pca_outer(evt);
    });
  }

  let pca_component_third = document.getElementById("bt-pca-component-third");
  if (pca_component_third) {
    pca_component_third.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        pca_eigen_rank = 2;
        set_pca_display_knobs();
        make_current_pca_outer(evt);
      }
    });
    pca_component_third.addEventListener("click", function (evt) {
      pca_eigen_rank = 2;
      set_pca_display_knobs();
      make_current_pca_outer(evt);
    });
  }

  let pca_display_absolute = document.getElementById("bt-pca-display-absolute");
  if (pca_display_absolute) {
    pca_display_absolute.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        pca_eigen_mode = 0;
        set_pca_display_knobs();
        make_current_pca_outer(evt);
      }
    });
    pca_display_absolute.addEventListener("click", function (evt) {
      pca_eigen_mode = 0;
      set_pca_display_knobs();
      make_current_pca_outer(evt);
    });
  }

  let pca_display_right = document.getElementById("bt-pca-display-right");
  if (pca_display_right) {
    pca_display_right.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        light_pca = 0;
        set_pca_display_light();
        set_pca_knobs_light();
        update_filters();
        pca_eigen_mode = 1;
        set_pca_display_knobs();
        make_current_pca_outer(evt);
      }
    });
    pca_display_right.addEventListener("click", function (evt) {
      light_pca = 0;
      set_pca_display_light();
      set_pca_knobs_light();
      update_filters();
      pca_eigen_mode = 1;
      set_pca_display_knobs();
      make_current_pca_outer(evt);
    });
  }

  let pca_display_left = document.getElementById("bt-pca-display-left");
  if (pca_display_left) {
    pca_display_left.addEventListener("keydown", function (evt) {
      evt = evt || window.event;
      if ((typeof evt.key !== "undefined") && (evt.key == "Enter")) {
        light_pca = 0;
        set_pca_display_light();
        set_pca_knobs_light();
        update_filters();
        pca_eigen_mode = -1;
        set_pca_display_knobs();
        make_current_pca_outer(evt);
      }
    });
    pca_display_left.addEventListener("click", function (evt) {
      light_pca = 0;
      set_pca_display_light();
      set_pca_knobs_light();
      update_filters();
      pca_eigen_mode = -1;
      set_pca_display_knobs();
      make_current_pca_outer(evt);
    });
  }

  pca_light_el = document.getElementById("rn-pca-light");
  if (pca_light_el) {
    pca_light_el.addEventListener("input", function (evt) {
      light_pca = parseInt(""+pca_light_el.value);
      set_pca_display_light();
      update_filters_outer(evt);
    });
  }
};

////////////////////////////////////////////////////////////////////////
// setting handlers for events to open local files

var setup_file_open = function () {
  debug_output("setup_file_open");

  var page_body_el = document.getElementById("page-body");
  var fl_outer_el = document.getElementById("file-parts");
  var fl_el = document.getElementById("in-load-file");

  if (page_body_el) {
    page_body_el.addEventListener("dragover", function (evt) {
      prevent_default(evt);
      if (evt) {
        evt.dataTransfer.dropEffect = 'copy';
      }
    }, false);
    page_body_el.addEventListener("drop", function (evt) {
      prevent_default(evt);
      if (evt && evt.dataTransfer && evt.dataTransfer.files && (evt.dataTransfer.files.length > 0)) {
        var file_dropped = evt.dataTransfer.files[0];
        if (file_dropped) {
          handle_local_image_file(file_dropped, true);
        }
      }
    }, false);
  }

  if (fl_outer_el) {
    fl_outer_el.addEventListener("dragover", function (evt) {
      prevent_default(evt);
      if (evt) {
        evt.dataTransfer.dropEffect = 'copy';
      }
    }, false);
    fl_outer_el.addEventListener("drop", function (evt) {
      prevent_default(evt);
      if (evt && evt.dataTransfer && evt.dataTransfer.files && (evt.dataTransfer.files.length > 0)) {
        var file_dropped = evt.dataTransfer.files[0];
        if (file_dropped) {
          handle_local_image_file(file_dropped, true);
        }
      }
    }, false);
  }

  if (fl_outer_el && fl_el) {
    fl_outer_el.addEventListener("click", function (evt) {
      prevent_default(evt);
      let click_event = new MouseEvent("click", {view: window, bubbles: false, cancelable: true});
      fl_el.dispatchEvent(click_event);
    });
  }

  if (fl_el) {
    fl_el.addEventListener("change", function (evt) {
      prevent_default(evt);
      var files = this.files;
      if (files.length > 0) {
        let img_file = this.files[0];
        if (img_file) {
          handle_local_image_file(img_file, false);
        }
      }
    }, false);
  }
};

var init_file_open_outer = function (evt) {
  prevent_default(evt);
  init_file_open();
};

var init_file_open = function () {
  var fl_el = document.getElementById("in-load-file");
  if (!fl_el) {
    return;
  }

  let click_event = new MouseEvent("click", {view: window, bubbles: false, cancelable: true});
  fl_el.dispatchEvent(click_event);
};

////////////////////////////////////////////////////////////////////////
// setting handlers for keyboard shortcuts

var set_key_actions = function () {

  document.addEventListener("keydown", function (evt) {
    evt = evt || window.event;

    if (!evt.ctrlKey) {
      return;
    }

    if (typeof evt.key !== "undefined") {
      switch (evt.key) {
        case "o":
          init_file_open_outer(evt);
          break;
        case ";":
          if (main_part in {"levels":true, "slopes":true, "ela":true, "pca":true}) {
            images_change_scaled(evt);
          }
          break;
        default:
          break;
      }
    }
  });

  document.addEventListener("keypress", function (evt) {
    evt = evt || window.event;

    if (!evt.ctrlKey) {
      return;
    }
    if (typeof evt.key !== "undefined") {
      switch (evt.key) {
        case "<":
          traverse_main_outer(evt, false);
          break;
        case ">":
          traverse_main_outer(evt, true);
          break;
        case ",":
          traverse_local_outer(evt, false);
          break;
        case ".":
          traverse_local_outer(evt, true);
          break;
        default:
          break;
      }
    }
  });

};

////////////////////////////////////////////////////////////////////////
// initial setting

var initialize_page_state = function () {

  can_ori_cvr_el = document.getElementById("visual-parts-cover-orig");
  can_res_cvr_el = document.getElementById("visual-parts-cover-result");
  can_thm_cvr_el = document.getElementById("visual-parts-cover-thumb");

  can_ori_el = document.getElementById("field-orig-image-canvas");
  can_res_levels_el = document.getElementById("field-result-levels-image-canvas");
  can_res_slopes_el = document.getElementById("field-result-slopes-image-canvas");
  can_res_ela_el = document.getElementById("field-result-ela-image-canvas");
  can_res_pca_el = document.getElementById("field-result-pca-image-canvas");
  can_aux_el = document.getElementById("field-aux-image-canvas");

  can_ori_scl_el = document.getElementById("field-orig-image-scaled-canvas");
  can_res_scl_el = document.getElementById("field-result-image-scaled-canvas");

  img_size_width_el = document.getElementById("info-sizes-width");
  img_size_height_el = document.getElementById("info-sizes-height");

  bt_meta_exif = document.getElementById("ctrl-meta-exif");
  bt_meta_thumb = document.getElementById("ctrl-meta-thumb");
  bt_meta_file = document.getElementById("ctrl-meta-file");

  setup_page_elements();
  images_show_stage(img_result_form);

  zero_ini_values();
  set_image_is_loaded(false);

  load_image_from_url();

  window.onresize = function (evt) {
    update_scaled_view_outer(evt);
  };

  setup_file_open();
  set_key_actions();
};

if ((document.readyState == "loading") || (document.readyState == "uninitialized")) {
  if (document.addEventListener) {
    document.addEventListener("DOMContentLoaded", initialize_page_state);
  } else {
    window.onload = initialize_page_state;
  }
} else {
  initialize_page_state();
}

